# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gi.repository import GObject, Gtk

class PanelView(Gtk.Bin, Gtk.Orientable):
    """
    Base class for a View
    """

    orientation = GObject.property(type=Gtk.Orientation, default=Gtk.Orientation.HORIZONTAL)

    def __init__(self, context):
        Gtk.Bin.__init__(self)
        self._context = context

    def _get_size(self, orientation):
        minimum, maximum = 0, 0
        child = self.get_child()

        if child is not None and child.get_visible():
            if orientation == Gtk.Orientation.HORIZONTAL:
                minimum, maximum = child.get_preferred_width()
            else:
                minimum, maximum = child.get_preferred_height()

        return minimum, maximum

    def do_get_preferred_width(self):
        return self._get_size(Gtk.Orientation.HORIZONTAL)

    def do_get_preferred_height(self):
        return self._get_size(Gtk.Orientation.VERTICAL)

    def do_size_allocate(self, allocation):
        Gtk.Bin.do_size_allocate(self, allocation)

        child = self.get_child()
        if child is not None and child.get_visible():
            child.size_allocate(allocation)

    # these should be overriden by subclasses

    # a label string used for this view
    def get_label(self):
        raise NotImplementedError

    # an icon for this view (Gtk.Image or a stock_id string)
    def get_icon(self):
        return None

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.outline
"""

from logging import getLogger

from .parser import Node
from ..issues import Issue


class OutlineNode(list):

    ROOT, STRUCTURE, LABEL, NEWCOMMAND, REFERENCE, GRAPHICS, PACKAGE, TABLE, NEWENVIRONMENT = list(range(9))

    def __init__(self, type, start=None, end=None, value=None, level=None, foreign=False, numOfArgs=None, file=None, **kwargs):
        """
        numOfArgs        only used for NEWCOMMAND type
        """
        self.type = type
        self.start = start
        self.end = end
        self.value = value
        self.level = level
        self.foreign = foreign
        self.numOfArgs = numOfArgs
        self.file = file

        self.oldcmd = kwargs.get("oldcmd")

    @property
    def xml(self):
        if self.type == self.ROOT:
            return "<root>%s</root>" % "".join([child.xml for child in self])
        elif self.type == self.STRUCTURE:
            return "<structure level=\"%s\" headline=\"%s\">%s</structure>" % (self.level, self.value,
                                                                                "".join([child.xml for child in self]))


class Outline(object):

    def __init__(self):
        self.rootNode = OutlineNode(OutlineNode.ROOT, level=0)
        self.labels = []            # OutlineNode objects
        self.bibliographies = []    # File objects
        self.colors = []
        self.packages = []           # OutlineNode objects
        self.newcommands = []        # OutlineNode objects
        self.newenvironments = []    # OutlineNode objects

        self.new_ref_commands = {}

    REF_CMDS = set(("ref","eqref","pageref"))
    def is_ref_command(self, cmd_name):
        return (cmd_name in self.REF_CMDS) or (cmd_name in self.new_ref_commands) 

from ..file import File
from ..preferences import Preferences


class LaTeXOutlineGenerator(object):

    _log = getLogger("LaTeXOutlineGenerator")

    # TODO: foreign flag is not necessary

    _STRUCTURE_LEVELS = { "part" : 1, "part*" : 1,
                          "chapter" : 2, "chapter*" : 2,
                          "section" : 3, "section*" : 3,
                          "subsection" : 4, "subsection*" : 4,
                          "subsubsection" : 5, "subsubsection*" : 5,
                          "paragraph" : 6,
                          "subparagraph" : 7 }

#    def __init__(self):
#        # TODO: read config
#        self.cfgLabelsInTree = True
#        self.cfgTablesInTree = True
#        self.cfgGraphicsInTree = True

    def generate(self, documentNode, issue_handler):
        """
        Generates an outline model from a document model and returns a list
        of list of issues if some occured.
        """

        # setup
        self.cfgLabelsInTree = Preferences().get("outline-show-labels")
        self.cfgTablesInTree = Preferences().get("outline-show-tables")
        self.cfgGraphicsInTree = Preferences().get("outline-show-graphics")

        self._outline = Outline()
        self._stack = [self._outline.rootNode]

        self._labelCache = {}

#        self._file = documentNode.value        # this is updated when a DOCUMENT occurs

        self._walk(documentNode, issue_handler)

        return self._outline

    def _walk(self, parentNode, issue_handler, foreign=False):
        """
        Recursively walk a node in the document model

        foreign        if True this node is a child of a reference node, so it's coming
                    from an expanded reference
        """

        childForeign = foreign

        for node in parentNode:
#            if node.type == Node.DOCUMENT:
#                self._file = node.value
            if node.type == Node.COMMAND:
                if node.value in list(self._STRUCTURE_LEVELS.keys()):
                    try:
                        headline = node.firstOfType(Node.MANDATORY_ARGUMENT).innerMarkup
                        level = self._STRUCTURE_LEVELS[node.value]
                        outlineNode = OutlineNode(OutlineNode.STRUCTURE, node.start, node.lastEnd, headline, level, foreign, file=node.file)

                        while self._stack[-1].level >= level:
                            self._stack.pop()

                        self._stack[-1].append(outlineNode)
                        self._stack.append(outlineNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed structure command", node.start, node.end, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "label":
                    try:
                        value = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText

                        if value in list(self._labelCache.keys()):
                            start, end = self._labelCache[value]
                            issue_handler.issue(Issue("Label <b>%s</b> has already been defined" % value, start, end, node.file, Issue.SEVERITY_ERROR))
                        else:
                            self._labelCache[value] = (node.start, node.lastEnd)

                            labelNode = OutlineNode(OutlineNode.LABEL, node.start, node.lastEnd, value, foreign=foreign, file=node.file)

                            self._outline.labels.append(labelNode)
                            if self.cfgLabelsInTree:
                                self._stack[-1].append(labelNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

#                elif node.value == "begin":
#                    environment = str(node.filter(Node.MANDATORY_ARGUMENT)[0][0])
#                    if environment == "lstlisting":
#                        # look for label in listing environment
#                        try:
#                            # TODO: Node should have a method like toDict() or something
#                            optionNode = node.filter(Node.OPTIONAL_ARGUMENT)[0]
#                            option = "".join([str(child) for child in optionNode])
#                            for pair in option.split(","):
#                                key, value = pair.split("=")
#                                if key.strip() == "label":
#                                    labelNode = OutlineNode(OutlineNode.LABEL, node.start, node.end, value.strip())
#                                    outline.labels.append(labelNode)
#                                    if self.cfgLabelsInTree:
#                                        stack[-1].append(labelNode)
#                        except IndexError:
#                            pass

                elif node.value == "usepackage":
                    try:
                        package = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        packageNode = OutlineNode(OutlineNode.PACKAGE, node.start, node.lastEnd, package, file=node.file)
                        self._outline.packages.append(packageNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.end, node.file, Issue.SEVERITY_ERROR))

                elif self.cfgTablesInTree and node.value == "begin":
                    try:
                        environ = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        if environ == "tabular":
                            tableNode = OutlineNode(OutlineNode.TABLE, node.start, node.lastEnd, "", foreign=foreign, file=node.file)
                            self._stack[-1].append(tableNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif self.cfgGraphicsInTree and node.value == "includegraphics":
                    try:
                        target = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        graphicsNode = OutlineNode(OutlineNode.GRAPHICS, node.start, node.lastEnd, target, foreign=foreign, file=node.file)
                        self._stack[-1].append(graphicsNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "bibliography":
                    try:
                        value = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        for bib in value.split(","):
                            self._outline.bibliographies.append(File("%s/%s.bib" % (node.file.dirname, bib)))
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "definecolor" or node.value == "xdefinecolor":
                    try:
                        name = str(node.firstOfType(Node.MANDATORY_ARGUMENT)[0])
                        self._outline.colors.append(name)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "newcommand":
                    try:
                        name = str(node.firstOfType(Node.MANDATORY_ARGUMENT)[0])[1:]    # remove "\"
                        try:
                            nArgs = int(node.filter(Node.OPTIONAL_ARGUMENT)[0].innerText)
                        except IndexError:
                            nArgs = 0
                        except Exception:
                            issue_handler.issue(Issue("Malformed newcommand", node.start, node.end, node.file, Issue.SEVERITY_ERROR))
                            nArgs = 0

                        #if the command has only one argument, be smart and see if it is a redefinition of an
                        #existing latex command
                        oldcmd = None
                        if nArgs == 1:
                            oldcommandnode = None
                            newcommandsnode = node.filter(Node.MANDATORY_ARGUMENT)[1]
                            #find the command that takes the '#1' argument
                            for i in newcommandsnode.filter(Node.COMMAND):
                                for j in i.filter(Node.MANDATORY_ARGUMENT):
                                    if j.innerText == "#1":
                                        oldcommandnode = i
                            if oldcommandnode:
                                oldcmd = i.value

                        ncNode = OutlineNode(OutlineNode.NEWCOMMAND, node.start, node.lastEnd, name, numOfArgs=nArgs, file=node.file, oldcmd=oldcmd)
                        self._outline.newcommands.append(ncNode)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                    # don't walk through \newcommand
                    continue

                elif node.value in ["newenvironment", "newtheorem"]:
                    try:
                        name = node.firstOfType(Node.MANDATORY_ARGUMENT).innerText
                        #try:
                        #    n_args = int(node.firstOfType(Node.OPTIONAL_ARGUMENT).innerText)
                        #except IndexError:
                        #    n_args = 0
                        ne_node = OutlineNode(OutlineNode.NEWENVIRONMENT, node.start, node.lastEnd, name, numOfArgs=0, file=node.file)
                        self._outline.newenvironments.append(ne_node)
                    except IndexError:
                        issue_handler.issue(Issue("Malformed command", node.start, node.lastEnd, node.file, Issue.SEVERITY_ERROR))

                elif node.value == "include" or node.value == "input":
                    childForeign = True

            self._walk(node, issue_handler, childForeign)


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.dialogs
"""

from gi.repository import Gtk

from ..util import GladeInterface
from ..resources import Resources
from .model import BibTeXModel


class InsertBibTeXEntryDialog(GladeInterface):

    _dialog = None

    def __init__(self):
        GladeInterface.__init__(self)
        self.filename = Resources().get_ui_file("insert_bibtex_entry.ui")

    def run(self):
        dialog = self._getDialog()

        result = None

        if dialog.run() == 1:

            s = "@%s{%s,\n" % (self._activeType.name, self._entryIdent.get_text())

            # fields
            f = []
            for entry, field in self._mapping.items():
                value = entry.get_text()
                if len(value):
                    if field == "title":
                        f.append("\t%s = {{%s}}" % (field, value))
                    else:
                        f.append("\t%s = {%s}" % (field, value))
            s += ",\n".join(f)

            # abstract
            #buf = self._view_abstract.get_buffer()
            #abs = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
            #if len(abs):
            #    s += ",\n\tabstract = {%s}" % abs

            s += "\n}"

            result = s

        dialog.hide()

        return result

    def _getDialog(self):
        if not self._dialog:
            self._dialog = self.find_widget("dialogInsertBibtexEntry")

            # get BibTeX model
            self._model = BibTeXModel()

            # setup types combobox

            self._activeType = None
            self._mapping = {}        # this maps Gtk.Entry objects to field names
            self._fieldCache = {}    # this is used to restore field values after the type has changed

            self._storeTypes = Gtk.ListStore(str)
            for t in self._model.types:
                self._storeTypes.append([t.name])

            comboTypes = self.find_widget("comboTypes")
            comboTypes.set_model(self._storeTypes)
            cell = Gtk.CellRendererText()
            comboTypes.pack_start(cell, True)
            comboTypes.add_attribute(cell, "text", 0)

            self._boxRequired = self.find_widget("boxRequired")
            self._boxOptional = self.find_widget("boxOptional")

            self._entryIdent = self.find_widget("entryIdent")

            self._buttonOk = self.find_widget("buttonOk")

            self.connect_signals({"on_comboTypes_changed": self._comboTypesChanged,
                                "on_entryIdent_changed": self._identChanged})

            comboTypes.set_active(0)

        return self._dialog

    def _identChanged(self, entry):
        enable = bool(len(entry.get_text()))
        self._buttonOk.set_sensitive(enable)

    def _comboTypesChanged(self, combo):
        i = combo.get_active()
        self._activeType = self._model.types[i]

        # cache values

        for entry, fieldName in self._mapping.items():
            text = entry.get_text()
            if len(text):
                self._fieldCache[fieldName] = entry.get_text()

        # reset mapping

        self._mapping = {}

        # required fields

        tbl_required = Gtk.Table()
        tbl_required.set_border_width(5)
        tbl_required.set_row_spacings(5)
        tbl_required.set_col_spacings(5)
        i = 0

        for field in self._activeType.required_fields:
            label = Gtk.Label(label=field.label + ":")
            label.set_alignment(0, .5)

            entry = Gtk.Entry()

            tbl_required.attach(label, 0, 1, i, i + 1, xoptions=Gtk.AttachOptions.FILL)
            tbl_required.attach(entry, 1, 2, i, i + 1)

            self._mapping[entry] = field.name

            # try to restore field value
            try:
                entry.set_text(self._fieldCache[field.name])
            except KeyError:
                pass

            i += 1

        for child in self._boxRequired.get_children():
            child.destroy()

        tbl_required.show_all()
        self._boxRequired.pack_start(tbl_required, False, False, 0)

        # optional fields

        tbl_optional = Gtk.Table()
        tbl_optional.set_border_width(5)
        tbl_optional.set_row_spacings(5)
        tbl_optional.set_col_spacings(5)
        i = 0

        for field in self._activeType.optional_fields:
            label = Gtk.Label(label=field.label + ":")
            label.set_alignment(0, .5)

            entry = Gtk.Entry()

            tbl_optional.attach(label, 0, 1, i, i + 1, xoptions=Gtk.AttachOptions.FILL)
            tbl_optional.attach(entry, 1, 2, i, i + 1)

            self._mapping[entry] = field.name

            # try to restore field value
            try:
                entry.set_text(self._fieldCache[field.name])
            except KeyError:
                pass

            i += 1

        for child in self._boxOptional.get_children():
            child.destroy()

        tbl_optional.show_all()
        self._boxOptional.pack_start(tbl_optional, False, False, 0)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.parser

BibTeX parser and object model
"""

#
##
## async parser feature
##
#if __name__ == "__main__":
#    #
#    # The script has been started in a shell process
#    #
#    # Parse the file passed as first argument and write the model
#    # as a pickled object to STDOUT
#    #
#    import sys
#
#    # TODO: fetch issues and pass them together with the model in a special
#    # transfer object
#
#    plugin_path = sys.argv[1]
#    filename = sys.argv[2]
#
#    sys.path.append(plugin_path)
#    sys.path.append("/home/michael/.gnome2/gedit/plugins")
#
#    from issues import MockIssueHandler
#    from base.file import File
#
#    model = BibTeXParser().parse_async(open(filename).read(), filename)
#else:
#    #
#    # normal package code...
#    #

from xml.sax.saxutils import escape

from ..issues import Issue
from ..preferences import Preferences

class Token(object):
    """
    A BibTeX token
    """

    AT, TEXT, COMMA, EQUALS, QUOTE, HASH, CURLY_OPEN, CURLY_CLOSE, ROUND_OPEN, ROUND_CLOSE = list(range(10))

    def __init__(self, type, offset, value):
        self.type = type
        self.offset = offset
        self.value = value

    def __str__(self):
        return "<Token type='%s' value='%s' @%s>" % (self.type, self.value, self.offset)


from ..util import StringReader
from ..util import open_info


class BibTeXLexer(object):
    """
    BibTeX lexer. We only separate text from special tokens here and
    apply escaping.
    """

    _TERMINALS_TOKENS = {"@" : Token.AT, "," : Token.COMMA, "=" : Token.EQUALS,
                         "{" : Token.CURLY_OPEN, "}" : Token.CURLY_CLOSE, "\"" : Token.QUOTE,
                         "#" : Token.HASH, "(" : Token.ROUND_OPEN, ")" : Token.ROUND_CLOSE}

    _TERMINALS = set(_TERMINALS_TOKENS.keys())

    def __init__(self, string):
        self._reader = StringReader(string)

    def __iter__(self):
        return self

    def __next__(self):
        """
        Return the next token
        """

        escaping = False
        textBuilder = None
        textStart = None

        while True:
            c = self._reader.read()

            if not escaping and c in self._TERMINALS:
                if textBuilder is not None:
                    self._reader.unread(c)
                    text = "".join(textBuilder)
                    textBuilder = None
                    return Token(Token.TEXT, textStart, text)

                return Token(self._TERMINALS_TOKENS[c], self._reader.offset - 1, c)

            else:
                if textBuilder is None:
                    textStart = self._reader.offset
                    textBuilder = [c]
                else:
                    textBuilder.append(c)

                if c == "\\":
                    escaping = True
                else:
                    escaping = False


class BibTeXParser(object):
    """
    A fast and safe BibTeX parser that generates a handy model on the fly.

    Instead of raising exceptions this parser uses an IIssueHandler.
    """

    _OUTSIDE, _TYPE, _AFTER_TYPE, _AFTER_STRING_TYPE, _KEY, _STRING_KEY, _AFTER_KEY, _AFTER_STRING_KEY, \
            _STRING_VALUE, _QUOTED_STRING_VALUE, _FIELD_NAME, _AFTER_FIELD_NAME, _FIELD_VALUE, _EMBRACED_FIELD_VALUE, \
            _QUOTED_FIELD_VALUE = list(range(15))

    def __init__(self, quiet=False):
        self._quiet = quiet
        self._max_size_info_shown = False

        self._state = None
        self._type = None
        self._constant = None
        self._entry = None
        self._file = None
        self._closingDelimiter = None
        self._document = None
        self._field = None
        self._value = None
        self._stack = None

    #
    # callables for each state of the parser
    #

    def _on_outside(self, token, file, issue_handler):
        if token.type == Token.AT:
            self._state = self._TYPE

    def _on_type(self, token, file, issue_handler):
        if token.type == Token.TEXT:
            self._type = token.value.strip()

            if self._type.lower() == "string" :
                self._constant = Constant()
                self._state = self._AFTER_STRING_TYPE
            elif self._type.lower() in ["preamble", "comment"]:        # simply skip PREAMBLE and COMMENT entries
                self._state = self._OUTSIDE
            else:
                self._entry = Entry()
                self._entry.type = self._type
                self._entry.start = token.offset - 2
                self._state = self._AFTER_TYPE
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> in entry type" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_after_type(self, token, file, issue_handler):
        if token.type == Token.CURLY_OPEN:
            self._closingDelimiter = Token.CURLY_CLOSE
            self._state = self._KEY
        elif token.type == Token.ROUND_OPEN:
            self._closingDelimiter = Token.ROUND_CLOSE
            self._state = self._KEY
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> after entry type" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_after_string_type(self, token, file, issue_handler):
        if token.type == Token.CURLY_OPEN:
            self._closingDelimiter = Token.CURLY_CLOSE
            self._state = self._STRING_KEY
        elif token.type == Token.ROUND_OPEN:
            self._closingDelimiter = Token.ROUND_CLOSE
            self._state = self._STRING_KEY
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> after string type" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._constant = None
            self._state = self._OUTSIDE

    def _on_key(self, token, file, issue_handler):
        if token.type == Token.TEXT:
            self._entry.key = token.value.strip()
            self._state = self._AFTER_KEY
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> in entry key" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_string_key(self, token, file, issue_handler):
        if token.type == Token.TEXT:
            self._constant.name = token.value.strip()
            self._state = self._AFTER_STRING_KEY
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> in string key" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._constant = None
            self._state = self._OUTSIDE

    def _on_after_key(self, token, file, issue_handler):
        if token.type == Token.COMMA:
            self._state = self._FIELD_NAME
        elif token.type == self._closingDelimiter:
            self._entry.end = token.offset + 1
            self._document.entries.append(self._entry)
            self._state = self._OUTSIDE
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> after entry key" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_after_string_key(self, token, file, issue_handler):
        if token.type == Token.EQUALS:
            self._state = self._STRING_VALUE
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> after string key" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._constant = None
            self._state = self._OUTSIDE

    def _on_string_value(self, token, file, issue_handler):
        if token.type == Token.QUOTE:
            self._state = self._QUOTED_STRING_VALUE

    def _on_quoted_string_value(self, token, file, issue_handler):
        if token.type == Token.TEXT:
            self._constant.value = token.value
            self._document.constants.append(self._constant)
            self._state = self._OUTSIDE

    def _on_field_name(self, token, file, issue_handler):
        if token.type == Token.TEXT:

            if token.value.isspace():
                return

            self._field = Field()
            self._field.name = token.value.strip()
            self._state = self._AFTER_FIELD_NAME
        elif token.type == self._closingDelimiter:
            self._entry.end = token.offset + 1
            self._document.entries.append(self._entry)
            self._state = self._OUTSIDE
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> in field name" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_after_field_name(self, token, file, issue_handler):
        if token.type == Token.EQUALS:
            self._state = self._FIELD_VALUE
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> after field name" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_field_value(self, token, file, issue_handler):
        # TODO: we may not recognize something like "author = ," as an error

        if token.value.isspace():
            return

        if token.type == Token.TEXT:
            self._value = token.value.strip()
            if self._value.isdigit():
                self._field.value.append(NumberValue(self._value))
            else:
                self._field.value.append(ConstantReferenceValue(self._value))
        elif token.type == Token.CURLY_OPEN:
            self._value = ""
            self._stack = [Token.CURLY_OPEN]
            self._state = self._EMBRACED_FIELD_VALUE
        elif token.type == Token.QUOTE:
            self._value = ""
            #stack = [Token.QUOTE]
            self._state = self._QUOTED_FIELD_VALUE
        elif token.type == Token.COMMA:
            self._entry.fields.append(self._field)
            self._state = self._FIELD_NAME
        elif token.type == self._closingDelimiter:
            self._entry.fields.append(self._field)
            self._entry.end = token.offset + 1
            self._document.entries.append(self._entry)
            self._state = self._OUTSIDE
        elif token.type == Token.HASH:
            pass
        else:
            issue_handler.issue(Issue("Unexpected token <b>%s</b> in field value" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
            self._entry = None
            self._state = self._OUTSIDE

    def _on_embraced_field_value(self, token, file, issue_handler):
        if token.type == Token.CURLY_OPEN:
            self._stack.append(Token.CURLY_OPEN)
            self._value += token.value
        elif token.type == Token.CURLY_CLOSE:
            try:
                while self._stack[-1] != Token.CURLY_OPEN:
                    self._stack.pop()
                self._stack.pop()

                if len(self._stack) == 0:
                    self._field.value.append(StringValue(self._value))
                    self._state = self._FIELD_VALUE
                else:
                    self._value += token.value

            except IndexError:
                issue_handler.issue(Issue("Unexpected token <b>%s</b> in field value" % escape(token.value),
                                    token.offset, token.offset + 1, file, Issue.SEVERITY_ERROR))
                self._entry = None
                self._state = self._OUTSIDE
        else:
            self._value += token.value

    def _on_quoted_field_value(self, token, file, issue_handler):
        if token.type == Token.QUOTE:
            self._field.value.append(StringValue(self._value))
            self._state = self._FIELD_VALUE
        else:
            self._value += token.value

    def parse(self, string, file, issue_handler):
        """
        Parse a BibTeX content
        @param string: the content to be parsed
        @param file: the File object containing the BibTeX
        @param issue_handler: an object implementing IIssueHandler
        """
        self._document = Document()

        # respect maximum BibTeX file size
        max_size_kb = int(Preferences().get("maximum-bibtex-size"))
        length = len(string)

        if length > max_size_kb * 1024:
            if not self._quiet and not self._max_size_info_shown:
                open_info("BibTeX file will not be parsed", "The maximum size of BibTeX files to parse is set to %s KB." % max_size_kb)
                self._max_size_info_shown = True
            return self._document

        # parse
        self._state = self._OUTSIDE

        #
        # use this hash table instead of endless if...elif statements
        #
        callables = {
                self._OUTSIDE : self._on_outside,
                self._TYPE : self._on_type,
                self._AFTER_TYPE : self._on_after_type,
                self._AFTER_STRING_TYPE : self._on_after_string_type,
                self._KEY : self._on_key,
                self._STRING_KEY : self._on_string_key,
                self._AFTER_KEY : self._on_after_key,
                self._AFTER_STRING_KEY : self._on_after_string_key,
                self._STRING_VALUE : self._on_string_value,
                self._QUOTED_STRING_VALUE : self._on_quoted_string_value,
                self._FIELD_NAME : self._on_field_name,
                self._AFTER_FIELD_NAME : self._on_after_field_name,
                self._FIELD_VALUE : self._on_field_value,
                self._EMBRACED_FIELD_VALUE : self._on_embraced_field_value,
                self._QUOTED_FIELD_VALUE : self._on_quoted_field_value
        }

        for token in BibTeXLexer(string):
            callables[self._state].__call__(token, file, issue_handler)

        return self._document

#
# BibTeX object model
#

class Value(object):
    def __init__(self, text):
        self.text = text

    MAX_MARKUP_LENGTH = 50

    @property
    def markup(self):
        text = self.text

        # remove braces
        if text.startswith("{{") and text.endswith("}}"):
            text = text[2:-2]
        elif text.startswith("{") and text.endswith("}"):
            text = text[1:-1]
        elif text.startswith("\\url{") and text.endswith("}"):
            text = text[5:-1]

        # truncate
        if len(text) > self.MAX_MARKUP_LENGTH:
            text = text[:self.MAX_MARKUP_LENGTH] + "..."

        # remove newlines
        text = text.replace("\n", "")

        # escape problematic characters
        text = escape(text)

        return text

    def __str__(self):
        return "<Value text='%s'>" % self.text


class NumberValue(Value):
    def __str__(self):
        return "<NumberValue text='%s'>" % self.text


class StringValue(Value):
    def __str__(self):
        return "<StringValue text='%s'>" % self.text


class ConstantReferenceValue(Value):
    @property
    def markup(self):
        return "<i>%s</i>" % escape(self.text)

    def __str__(self):
        return "<ReferenceValue text='%s'>" % self.text


class Field(object):
    def __init__(self):
        self.name = None
        self.value = []

    @property
    def valueMarkup(self):
        return " ".join([v.markup for v in self.value])

    @property
    def valueString(self):
        return " ".join([v.text for v in self.value])

    def __str__(self):
        return "<Field name='%s' value='%s' />" % (self.name, self.valueString)


class Entry(object):
    def __init__(self):
        self.type = None
        self.key = None
        self.start = None
        self.end = None
        self.fields = []

    def findField(self, name):
        for field in self.fields:
            if field.name == name:
                return field
        raise KeyError

    def __str__(self):
        s = "<Entry type='%s' key='%s'>\n" % (self.type, self.key)
        for field in self.fields:
            s += "\t" + str(field) + "\n"
        s += "</Entry>"
        return s


class Constant(object):
    """
    A BibTeX string constant
    """
    def __init__(self):
        self.name = None
        self.value = None


class Document(object):
    def __init__(self):
        self.entries = []
        self.constants = []

    def __str__(self):
        s = "<Document>\n"
        for entry in self.entries:
            s += str(entry) + "\n"
        s += "</Document>"
        return s


#    #
#    # async parser feature
#    #
#
#    # TODO: put the __main__ part in another file
#
#    import pickle
#    import os
#
#    from ..tools.util import Process
#    from ..base.resources import PLUGIN_PATH
#
#    # TODO: time pickle.loads() and pickle.dump()
#    # TODO: support Process.abort()
#
#    class AsyncParserRunner(Process):
#
#        __log = getLogger("AsyncParserRunner")
#
#        def parse(self, file):
#            self.__pickled_object = None
#
#            source_path = PLUGIN_PATH + "/src"
#            self.__log.debug("chdir: %s" % source_path)
#            os.chdir(source_path)
#
#            self.execute("python %s/bibtex/parser.py %s %s" % (source_path, source_path, file.path))
#
#        def _on_stdout(self, text):
#            # Process._on_stdout
#            self.__pickled_object = text
#
#        def _on_stderr(self, text):
#            # Process._on_stderr
#            self.__log.debug("_on_stderr: %s" % text)
#
#        def _on_abort(self):
#            # Process._on_abort
#            pass
#
#        def _on_exit(self, condition):
#            # Process._on_exit
#            self.__log.debug("_on_exit")
#
#            model = None
#
#            if condition:
#                self.__log.error("failed")
#            else:
#                model = pickle.loads(self.__pickled_object)
#
#            self._on_parser_finished(model)
#
#        def _on_parser_finished(self, model):
#            """
#            To be overridden by the subclass
#            """






# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.views

LaTeX-specific views
"""

import logging
import xml.etree.ElementTree as ElementTree

from gi.repository import Gtk, GdkPixbuf
from os.path import basename

from ..preferences import Preferences
from ..panelview import PanelView
from ..file import File
from ..resources import Resources
from ..snippetmanager import SnippetManager
from ..outline import OutlineOffsetMap, BaseOutlineView
from .outline import OutlineNode
from ..gldefs import _

LOG = logging.getLogger(__name__)

class SymbolCollection(object):
    """
    A collection of symbols read from an XML file
    """

    class Group(object):
        def __init__(self, label):
            """
            @param label: a label for this Group
            """
            self.label = label
            self.symbols = []


    class Symbol(object):
        def __init__(self, snippet, icon):
            """
            @param snippet: a snippet to insert
            @param icon: an icon filename
            """
            self.snippet = snippet
            self.icon = icon


    def __init__(self):
        filename = Resources().get_data_file("symbols.xml")

        self.groups = []

        symbols_el = ElementTree.parse(filename).getroot()
        for group_el in symbols_el.findall("group"):
            group = self.Group(group_el.get("label"))
            for symbol_el in group_el.findall("symbol"):
                symbol = self.Symbol(symbol_el.text.strip(), Resources().get_icon("%s" % symbol_el.get("icon")))
                group.symbols.append(symbol)
            self.groups.append(group)


class LaTeXSymbolMapView(PanelView):
    """
    """

    def __init__(self, context, editor):
        PanelView.__init__(self, context)
        self._preferences = Preferences()

        self.set_orientation(Gtk.Orientation.VERTICAL)

        self._grid = Gtk.Grid()
        self._grid.set_orientation(Gtk.Orientation.VERTICAL)

        self._load_collection(SymbolCollection())

        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled.set_shadow_type(Gtk.ShadowType.NONE)
        scrolled.add_with_viewport(self._grid)
        scrolled.set_vexpand(True)

        self.add(scrolled)
        self.show_all()

    def get_label(self):
        return _("Symbols")

    def get_icon(self):
        return Gtk.Image.new_from_stock(Gtk.STOCK_INDEX,Gtk.IconSize.MENU)

    def _load_collection(self, collection):
        self._expanded_groups = set(self._preferences.get("expanded-symbol-groups").split(","))

        for group in collection.groups:
            self._add_group(group)

    def _add_group(self, group):
        model = Gtk.ListStore(GdkPixbuf.Pixbuf, str)        # icon, snippet

        for symbol in group.symbols:
            try:
                model.append([GdkPixbuf.Pixbuf.new_from_file(symbol.icon), symbol.snippet])
            except:
                LOG.error("Could not add symbol group %s to model" % symbol, exc_info=True)

        view = Gtk.IconView(model=model)
        view.set_pixbuf_column(0)
        view.set_selection_mode(Gtk.SelectionMode.SINGLE)
        view.connect("item-activated", self._on_symbol_activated)
        view.connect("focus-out-event", self._on_focus_out_event)
        view.set_spacing(0)
        view.set_column_spacing(0)
        view.set_row_spacing(0)
        view.set_columns(4)
        view.set_text_column(-1)
        view.set_tooltip_column(1)
        view.set_hexpand(True)
        view.set_halign(Gtk.Align.FILL)

        expander = Gtk.Expander(label=group.label)
        expander.set_hexpand(True)
        expander.set_halign(Gtk.Align.FILL)
        expander.add(view)
        if group.label in self._expanded_groups:
            expander.set_expanded(True)
        expander.connect("notify::expanded", self._on_group_expanded, group.label)

        self._grid.add(expander)

    def _on_group_expanded(self, expander, paramSpec, group_label):
        """
        The Expander for a symbol group has been expanded
        """
        if expander.get_expanded():
            self._expanded_groups.add(group_label)
        else:
            self._expanded_groups.remove(group_label)

        self._preferences.set("expanded-symbol-groups", ",".join(self._expanded_groups))

    def _on_symbol_activated(self, icon_view, path):
        """
        A symbol has been activated

        @param icon_view: the Gtk.IconView
        @param path: the Gtk.TreePath to the item
        """
        snippet = icon_view.get_model()[path][1]
        SnippetManager().insert_at_cursor(self._context.active_editor, snippet)

    def _on_focus_out_event(self, icon_view, event):
        icon_view.unselect_all()


class LaTeXOutlineView(BaseOutlineView):
    """
    A View showing an outline of the edited LaTeX document
    """

    def __init__(self, context, editor):
        BaseOutlineView.__init__(self, context, editor)
        self._handlers = {}

        self._offset_map = OutlineOffsetMap()

        # additional toolbar buttons
        btn_graphics = Gtk.ToggleToolButton()
        btn_graphics.set_icon_widget(Gtk.Image.new_from_file(Resources().get_icon("tree_includegraphics.png")))
        btn_graphics.set_tooltip_text(_("Show graphics"))
        self._toolbar.insert(btn_graphics, -1)

        btn_tables = Gtk.ToggleToolButton()
        btn_tables.set_icon_widget(Gtk.Image.new_from_file(Resources().get_icon("tree_table.png")))
        btn_tables.set_tooltip_text(_("Show tables"))
        self._toolbar.insert(btn_tables, -1)

        btn_graphics.set_active(Preferences().get("outline-show-graphics"))
        btn_tables.set_active(Preferences().get("outline-show-tables"))

        self._handlers[btn_graphics] = btn_graphics.connect("toggled", self._on_graphics_toggled)
        self._handlers[btn_tables] = btn_tables.connect("toggled", self._on_tables_toggled)

        self.show_all()

    def set_outline(self, outline):
        """
        Load a new outline model
        """
        LOG.debug("LatexOutline: set outline")

        self._save_state()

        self._offset_map = OutlineOffsetMap()
        OutlineConverter().convert(self._store, outline, self._offset_map, self._editor.edited_file)

        self._restore_state()

    def _on_node_activated(self, node):
        """
        An outline node has been double-clicked on

        @param node: an instance of latex.outline.OutlineNode
        """
        if node.type == OutlineNode.GRAPHICS:
            # use 'gvfs-open' to open the graphics file

            target = node.value

            if not target:
                return

            # the way we use a mixture of File objects and filenames is not optimal...
            if File.is_absolute(target):
                filename = target
            else:
                filename = File.create_from_relative_path(target, node.file.dirname).path

            # an image may be specified without the extension
            potential_extensions = ["", ".eps", ".pdf", ".jpg", ".jpeg", ".gif", ".png"]

            found = False
            for ext in potential_extensions:
                f = File(filename + ext)
                if f.exists:
                    found = True
                    break

            if not found:
                LOG.error("LatexOutline: File not found: %s" % filename)
                return

            Gtk.show_uri(None, f.uri, Gtk.get_current_event_time())

        else:
            # select/open/activate the referenced file
            if node.file == self._editor.edited_file:
                self._editor.select(node.start, node.end)
            else:
                self._context.activate_editor(node.file)

    def _on_tables_toggled(self, toggle_button):
        value = toggle_button.get_active()
#        Settings().set("LatexOutlineTables", value)
#        self.trigger("tablesToggled", value)
        Preferences().set("outline-show-tables", value)

    def _on_graphics_toggled(self, toggle_button):
        value = toggle_button.get_active()
#        Settings().set("LatexOutlineGraphics", value)
#        self.trigger("graphicsToggled", value)
        Preferences().set("outline-show-graphics", value)


class OutlineConverter(object):
    """
    This creates a Gtk.TreeStore object from a LaTeX outline model
    """

    def __init__(self):
        self._preferences = Preferences()
        self._ICON_LABEL = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("label.png"))
        self._ICON_TABLE = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_table.png"))
        self._ICON_GRAPHICS = GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_includegraphics.png"))

        self._LEVEL_ICONS = { 1 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_part.png")),
                    2 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_chapter.png")),
                    3 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_section.png")),
                    4 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_subsection.png")),
                    5 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_subsubsection.png")),
                    6 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_paragraph.png")),
                    7 : GdkPixbuf.Pixbuf.new_from_file(Resources().get_icon("tree_paragraph.png")) }

    def convert(self, tree_store, outline, offset_map, file):
        """
        Convert an Outline object to a Gtk.TreeStore and update an OutlineOffsetMap

        @param tree_store: Gtk.TreeStore
        @param outline: latex.outline.Outline
        @param offset_map: outline.OutlineOffsetMap
        @param file: the edited File (to identify foreign outline nodes)
        """
        self._offsetMap = offset_map
        self._treeStore = tree_store
        self._treeStore.clear()
        self._file = file

        self._append(None, outline.rootNode)

    def _append(self, parent, node):
        """
        Recursively append an outline node to the TreeStore

        @param parent: a Gtk.TreeIter or None
        @param node: an OutlineNode
        """
        value = node.value

        color = self._preferences.get("light-foreground-color")

        if node.file and node.file != self._file:
            value = "%s <span color='%s'>%s</span>" % (value, color, node.file.shortbasename)

        if node.type == OutlineNode.STRUCTURE:
            icon = self._LEVEL_ICONS[node.level]
            parent = self._treeStore.append(parent, [value, icon, node])
        elif node.type == OutlineNode.LABEL:
            parent = self._treeStore.append(parent, [value, self._ICON_LABEL, node])
        elif node.type == OutlineNode.TABLE:
            parent = self._treeStore.append(parent, [value, self._ICON_TABLE, node])
        elif node.type == OutlineNode.GRAPHICS:
            label = basename(node.value)

            if node.file and node.file != self._file:
                label = "%s <span color='%s'>%s</span>" % (label, color, node.file.shortbasename)

            parent = self._treeStore.append(parent, [label, self._ICON_GRAPHICS, node])

        # store path in offset map for all non-foreign nodes
        # check for parent to ignore root node
        if parent and not node.foreign:
            path = self._treeStore.get_path(parent)
            self._offsetMap.put(node.start, path)

        for child in node:
            self._append(parent, child)

# ex:ts=4:et:



#    Gedit latex plugin
#    Copyright (C) 2011 Ignacio Casal Quinteiro
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os.path
import platform

from gi.repository import GLib, Gedit, GObject, Gio, Gtk
from .resources import Resources

from .config import MENUACTIONS

from .preferences.tools import ToolPreferences

class LaTeXAppActivatable(GObject.Object, Gedit.AppActivatable):
    __gtype_name__ = "GeditLaTeXAppActivatable"

    app = GObject.property(type=Gedit.App)
    
    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        if platform.platform() == 'Windows':
            userdir = os.path.expanduser('~/gedit/latex')
        else:
            userdir = os.path.join(GLib.get_user_config_dir(), 'gedit/latex')

        #check if running from srcdir and if so, prefer that for all data files
        me = os.path.realpath(os.path.dirname(__file__))
        if os.path.exists(os.path.join(me, "..", "configure.ac")):
            sysdir = os.path.abspath(os.path.join(me, "..", "data"))
        else:
            sysdir = self.plugin_info.get_data_dir()

        Resources().set_dirs(userdir, sysdir)
        
        #The following is needed to support gedit 3.12 new menu api.
        #It adds menus and shortcuts here.
        #Actions and toolbar construction are still done in windowactivatable.py.
        
        self._tool_preferences = ToolPreferences()
        self._tool_preferences.connect("tools-changed", self._on_tools_changed)
        
        self.add_latex_menu()
        self.add_latex_tools_menu()
        self.init_tools()
        
    def add_latex_menu(self):
        self.appmenu_ext = self.extend_menu("preferences-section")
        
        self.menu_ext = self.extend_menu("tools-section-1")
        menu = Gio.MenuItem.new(_("LaTeX"), "win.FileDummyAction")
        container = Gio.Menu.new()
        menu.set_submenu(container)
        self.menu_ext.append_menu_item(menu)
        
        self._icon_factory = Gtk.IconFactory()
        self._icon_factory.add_default()
        
        for clazz in MENUACTIONS:
            action = clazz(icon_factory=self._icon_factory)
            actionlink = "win." + clazz.__name__
            item = Gio.MenuItem.new(_(action.label), actionlink)
            if clazz.__name__ == 'LaTeXNewAction':
                self.appmenu_ext.append_menu_item(item)
            else:
                container.append_item(item)
            # FIXME: this is not working (it does work in init_tools() below):
            item.set_attribute_value("hidden-when",
                                    GLib.Variant.new_string("action-disabled"))
            if action.accelerator is not None:
                self.app.add_accelerator(action.accelerator, actionlink, None)

    def add_latex_tools_menu(self):
        menu = Gio.MenuItem.new(_("LaTeX Tools"), "win.ToolsDummyAction")
        container = Gio.Menu()
        menu.set_submenu(container)
        self.latex_tools_menu = container
        menu.set_attribute_value("hidden-when",
                                    GLib.Variant.new_string("action-disabled"))
        self.menu_ext.append_menu_item(menu)
    
    def init_tools(self):
      
        #the following is copied from windowactivatable.py and modified as necessary
        
        i = 1                    # counting tool actions
        accel_counter = 1        # counting tool actions without custom accel
        for tool in self._tool_preferences.tools:
            # hopefully unique action name
            name = "Tool%sAction" % i

            actionlink = "win." + name
            item = Gio.MenuItem.new(_(tool.label), actionlink)
            item.set_attribute_value("hidden-when",
                                    GLib.Variant.new_string("action-disabled"))
            self.latex_tools_menu.append_item(item)

            accelerator = None
            if tool.accelerator and len(tool.accelerator) > 0:
                key,mods = Gtk.accelerator_parse(tool.accelerator)
                if Gtk.accelerator_valid(key,mods):
                    accelerator = tool.accelerator
            if not accelerator:
                accelerator = "<Ctrl><Alt>%s" % accel_counter
                accel_counter += 1
            self.app.add_accelerator(accelerator, actionlink, None)
            i += 1
            
    def _on_tools_changed(self, tools):
        self.latex_tools_menu.remove_all()
        self.init_tools()

    def do_deactivate(self):
        del self.latex_tools_menu
        del self.menu_ext
        del self.appmenu_ext
        
# ex:ts=4:et



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import logging
import uuid
from xml.etree import ElementTree
import os

from gi.repository import GObject

from ..resources import Resources
from ..tools import Tool, Job
from ..tools.postprocess import GenericPostProcessor, RubberPostProcessor, LaTeXPostProcessor
from ..util import singleton, open_error

LOG = logging.getLogger(__name__)

def str_to_bool(x):
    """
    Converts a string to a boolean value
    """
    if type(x) is bool:
        return x
    elif type(x) is str or type(x) is str:
        try:
            return {"false" : False, "0" : False, "true" : True, "1" : True}[x.strip().lower()]
        except KeyError:
            LOG.error("str_to_bool: unsupported value %s" % x)
    else:
        LOG.error("str_to_bool: unsupported type %s" % type(x))

@singleton
class ToolPreferences(GObject.GObject):

    __gsignals__ = {
        "tools-changed": (
            GObject.SignalFlags.RUN_LAST, None, []),
    }

    # maps names to classes
    POST_PROCESSORS = {"GenericPostProcessor" : GenericPostProcessor,
                       "LaTeXPostProcessor" : LaTeXPostProcessor,
                       "RubberPostProcessor" : RubberPostProcessor}

    def __init__(self):
        GObject.GObject.__init__(self)
        self.__tool_objects = None
        self.__tool_ids = None
        self.__tools_changed = False

        filename = Resources().get_user_file("tools.xml")
        try:
            self.__tools = ElementTree.parse(filename).getroot()
            LOG.debug("Found %s" % filename)
        except (ElementTree.ParseError, FileNotFoundError) as exc:
            if isinstance(exc, ElementTree.ParseError):
                i = 0
                while True:
                    destname = "%s.%d" % (filename, i)
                    if not os.path.exists(destname):
                        break
                    i += 1
                os.rename(filename, destname)
                open_error("The file \"%s\" is corrupted and cannot be "
                           "parsed." % filename,
                           "It was moved to \"%s\", and the Latex Plugin will "
                           "now fallback to the default tools." % destname)

            filename = Resources().get_data_file("tools.xml")

        self.__tools = ElementTree.parse(filename).getroot()
        LOG.debug("ToolPreferences constructed")

    def __notify_tools_changed(self):
        self.emit("tools-changed")

    @property
    def tools(self):
        """
        Return all Tools
        """
        self.__tool_ids = {}

        tools = []

        for tool_element in self.__tools.findall("tool"):
            jobs = []
            for job_element in tool_element.findall("job"):
                command = '' if job_element.text is None else job_element.text.strip()
                jobs.append(Job(command, str_to_bool(job_element.get("mustSucceed")), self.POST_PROCESSORS[job_element.get("postProcessor")]))

            assert not tool_element.get("extensions") is None

            extensions = tool_element.get("extensions").split()
            accelerator = tool_element.get("accelerator")
            id = tool_element.get("id")
            tool = Tool(tool_element.get("label"), jobs, tool_element.get("description"), accelerator, extensions)
            self.__tool_ids[tool] = id

            tools.append(tool)

        return tools

    def __find_tool_element(self, id):
        """
        Find the tool element with the given id
        """
        for element in self.__tools.findall("tool"):
            if element.get("id") == id:
                return element
        LOG.warning("<tool id='%s'> not found" % id)
        return None

    def save_or_update_tool(self, tool):
        """
        Save or update the XML subtree for the given Tool

        @param tool: a Tool object
        """
        tool_element = None
        if tool in self.__tool_ids:
            # find tool tag
            LOG.debug("Tool element found, updating...")

            id = self.__tool_ids[tool]
            tool_element = self.__find_tool_element(id)
        else:
            # create new tool tag
            LOG.debug("Creating new Tool...")

            id = str(uuid.uuid4())
            self.__tool_ids[tool] = id

            tool_element = ElementTree.SubElement(self.__tools, "tool")
            tool_element.set("id", id)
        
        tool_element.set("label", tool.label)
        tool_element.set("description", tool.description)
        tool_element.set("extensions", " ".join(tool.extensions))
        if tool.accelerator is None:
            if "accelerator" in list(tool_element.attrib.keys()):
                del tool_element.attrib["accelerator"]
        else:
            tool_element.set("accelerator", tool.accelerator)

        # remove all jobs
        for job_element in tool_element.findall("job"):
            tool_element.remove(job_element)

        # append new jobs
        for job in tool.jobs:
            job_element = ElementTree.SubElement(tool_element, "job")
            job_element.set("mustSucceed", str(job.must_succeed))
            job_element.set("postProcessor", job.post_processor.name)
            job_element.text = job.command_template

        self.__tools_changed = True
        self.__notify_tools_changed()

    def swap_tools(self, tool_1, tool_2):
        """
        Swap the order of two Tools
        """
        # grab their ids
        id_1 = self.__tool_ids[tool_1]
        id_2 = self.__tool_ids[tool_2]

        if id_1 == id_2:
            LOG.warning("Two tools have the same id. Please modify tools.xml to have unique id's.")
            return

        LOG.debug("Tool IDs are {%s: %s, %s, %s}" % (tool_1.label, id_1, tool_2.label, id_2))

        tool_element_1 = None
        tool_element_2 = None

        # find the XML elements and current indexes of the tools
        i = 0
        for tool_element in self.__tools:
            if tool_element.get("id") == id_1:
                tool_element_1 = tool_element
                index_1 = i
            elif tool_element.get("id") == id_2:
                tool_element_2 = tool_element
                index_2 = i

            if not (tool_element_1 is None or tool_element_2 is None):
                break

            i += 1

        LOG.debug("Found XML elements, indexes are {%s: %s, %s, %s}" % (tool_1.label, index_1, tool_2.label, index_2))

        # successively replace each of them by the other in the XML model
        self.__tools.remove(tool_element_1)
        self.__tools.insert(index_1, tool_element_2)

        LOG.debug("Replaced first tool by second in list")

        self.__tools.remove(tool_element_2)
        self.__tools.insert(index_2, tool_element_1)

        LOG.debug("Replaced second tool by first in list")

        # notify changes
        self.__tools_changed = True
        self.__notify_tools_changed()

    def delete_tool(self, tool):
        """
        Delete the given Tool

        @param tool: a Tool object
        """
        try:
            id = self.__tool_ids[tool]
            tool_element = self.__find_tool_element(id)
            self.__tools.remove(tool_element)

            del self.__tool_ids[tool]

            self.__tools_changed = True
        except KeyError as e:
            LOG.error("delete_tool: %s" % e)

        self.__notify_tools_changed()

    def save(self):
        """
        Save the preferences to XML
        """
        if self.__tools_changed:
            LOG.debug("Saving tools...")

            tree = ElementTree.ElementTree(self.__tools)
            tree.write(Resources().get_user_file("tools.xml"), encoding="utf-8")

            self.__tools_changed = False


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
tools.util
"""

import logging
import os
import signal
import subprocess
import fcntl

from gi.repository import GObject

LOG = logging.getLogger(__name__)

class Process(object):
    """
    This runs a command in a child process and polls the output
    """

    # intervall of polling stdout of the child process
    __POLL_INTERVAL = 250

    def execute(self, command):
        LOG.debug("execute: %s" % command)

        # run child process
        self.__process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)

        # unblock pipes
        fcntl.fcntl(self.__process.stdout, fcntl.F_SETFL, os.O_NONBLOCK)
        fcntl.fcntl(self.__process.stderr, fcntl.F_SETFL, os.O_NONBLOCK)

        # monitor process and pipes
        self.__handlers = [ GObject.timeout_add(self.__POLL_INTERVAL, self.__on_stdout),
                            GObject.timeout_add(self.__POLL_INTERVAL, self.__on_stderr),
                            GObject.child_watch_add(self.__process.pid, self.__on_exit) ]

    def abort(self):
        """
        Abort the running process
        """
        if self.__process:
            for handler in self.__handlers:
                GObject.source_remove(handler)

            try:
                os.kill(self.__process.pid, signal.SIGTERM)

                self._on_abort()
            except OSError as e:
                LOG.error("Failed to abort process: %s" % e)

    def __on_stdout(self):
        try:
            s = self.__process.stdout.read()
            if s != None:
                self._on_stdout(s.decode())
        except IOError:
            pass
        return True

    def __on_stderr(self):
        try:
            s = self.__process.stderr.read()
            if s != None:
                self._on_stderr(s.decode())
        except IOError:
            pass
        return True

    def __on_exit(self, pid, condition):
        for handler in self.__handlers:
            GObject.source_remove(handler)

        # read remaining output
        self.__on_stdout()
        self.__on_stderr()

        self._on_exit(condition)

    def _on_stdout(self, text):
        """
        To be overridden
        """

    def _on_stderr(self, text):
        """
        To be overridden
        """

    def _on_abort(self):
        """
        To be overridden
        """

    def _on_exit(self, condition):
        """
        To be overridden
        """


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.parser

LaTeX parser and object model
"""

from re import compile

from ..util import verbose, escape
from ..issues import Issue


class Node(list):
    """
    This is the base class of the LaTeX object model
    """

    DOCUMENT, COMMAND, MANDATORY_ARGUMENT, OPTIONAL_ARGUMENT, TEXT, EMBRACED = list(range(6))

    def __init__(self, type, value=None):
        self.type = type
        self.value = value
        self.parent = None

        # this indicates if an argument is closed or not
        # (only used by the PrefixParser)
        self.closed = False

    def firstOfType(self, type):
        """
        Return the first child node of a given type
        """
        for node in self:
            if node.type == type:
                return node
        raise IndexError

    def filter(self, type):
        """
        Return all child nodes of this node having a certain type
        """
        return [node for node in self if node.type == type]

    @property
    def xml(self):
        """
        Return an XML representation of this node (for debugging)
        """
        if self.type == self.COMMAND:
            content = "".join([node.xml for node in self])
            if len(content):
                return "<command name=\"%s\">%s</command>" % (escape(self.value), content)
            else:
                return "<command name=\"%s\" />" % escape(self.value)
        elif self.type == self.MANDATORY_ARGUMENT:
            return "<mandatory>%s</mandatory>" % "".join([node.xml for node in self])
        elif self.type == self.OPTIONAL_ARGUMENT:
            return "<optional>%s</optional>" % "".join([node.xml for node in self])
        elif self.type == self.TEXT:
            return escape(self.value)
        elif self.type == self.DOCUMENT:
            return "<document>%s</document>" % "".join([node.xml for node in self])
        elif self.type == self.EMBRACED:
            return "<embraced>%s</embraced>" % "".join([node.xml for node in self])

    @property
    def xmlPrefix(self):
        """
        Return an XML representation of this node (for debugging)

        This is for the prefix mode, so we print if the arguments are closed or not
        """
        if self.type == self.COMMAND:
            content = "".join([node.xmlPrefix for node in self])
            if len(content):
                return "<command name=\"%s\">%s</command>" % (self.value, content)
            else:
                return "<command name=\"%s\" />" % self.value
        elif self.type == self.MANDATORY_ARGUMENT:
            return "<mandatory closed=%s>%s</mandatory>" % (self.closed, "".join([node.xmlPrefix for node in self]))
        elif self.type == self.OPTIONAL_ARGUMENT:
            return "<optional closed=%s>%s</optional>" % (self.closed, "".join([node.xmlPrefix for node in self]))
        elif self.type == self.TEXT:
            return escape(self.value)
        elif self.type == self.DOCUMENT:
            return "<document>%s</document>" % "".join([node.xmlPrefix for node in self])
        elif self.type == self.EMBRACED:
            return "<embraced>%s</embraced>" % "".join([node.xmlPrefix for node in self])

    def __str__(self):
        """
        Return the original LaTeX representation of this node
        """
        if self.type == self.COMMAND:
            return "\\%s%s" % (self.value, "".join([str(node) for node in self]))
        elif self.type == self.MANDATORY_ARGUMENT or self.type == self.EMBRACED:
            return "{%s}" % "".join([str(node) for node in self])
        elif self.type == self.OPTIONAL_ARGUMENT:
            return "[%s]" % "".join([str(node) for node in self])
        elif self.type == self.TEXT:
            return self.value
        elif self.type == self.DOCUMENT:
            return "".join([str(node) for node in self])

    @property
    def innerText(self):
        """
        Return the concatenated values of all TEXT child nodes
        """
        return "".join([child.value for child in self if child.type == Node.TEXT])

    @property
    def markup(self):
        """
        Return the concatenated markup values of this node and all child nodes
        """
        if self.type == self.COMMAND:
            return "<span color=\"grey\">\\%s</span>%s" % (escape(self.value), "".join([node.markup for node in self]))
        elif self.type == self.MANDATORY_ARGUMENT or self.type == self.EMBRACED:
            return "<span color=\"grey\">{</span>%s<span color=\"grey\">}</span>" % "".join([node.markup for node in self])
        elif self.type == self.OPTIONAL_ARGUMENT:
            return "<span color=\"grey\">[</span>%s<span color=\"grey\">]</span>" % "".join([node.markup for node in self])
        elif self.type == self.TEXT:
            return escape(self.value)
        elif self.type == self.DOCUMENT:
            return "".join([node.markup for node in self])

    @property
    def innerMarkup(self):
        """
        Return the concatenated markup values of only the child nodes
        """
        return "".join([node.markup for node in self])

    def append(self, node):
        """
        Append a child node and store a back-reference
        """
        node.parent = self
        list.append(self, node)

    def find(self, value):
        """
        Find child node with given value (recursive, so grand-children are found, too)
        """
        # TODO

    def destroy(self):
        # This is important, python cannot automatically free this
        # object because of cyclic references (it is doubly linked)
        self.parent = None
        for child in self:
            child.destroy()
        del self[:]

    #~ def __del__(self):
        #~ print "Properly destroyed %s" % self



class Document(Node):
    """
    An extended Node with special methods for a LaTeX document
    """

    # FIXME: doesn't extend LocalizedNode which leads to
    # https://sourceforge.net/tracker/index.php?func=detail&aid=2899795&group_id=204144&atid=988428

    # TODO: implement a generic interface for searching the document model

    def __init__(self, file):
        Node.__init__(self, Node.DOCUMENT, file)

        self._is_master_called = False
        self._is_master = False

        self._end_of_document = None
        self._end_of_packages = None

    def _do_is_master(self):
        # TODO: this should be recursive

        for node in self:
            if node.type == Node.COMMAND and node.value == "begin":
                if node.firstOfType(Node.MANDATORY_ARGUMENT).innerText == "document":
                    return True
        return False

    @property
    def is_master(self):
        """
        @return: True if this document is a master document
        """

        # TODO: determine this while parsing

        if not self._is_master_called:
            self._is_master = self._do_is_master()
            self._is_master_called = True
        return self._is_master

    def _find_end_of_document(self):
        # TODO: this should be recursive

        for node in self:
            if node.type == Node.COMMAND and node.value == "end":
                if node.firstOfType(Node.MANDATORY_ARGUMENT).innerText == "document":
                    return node.start
        return 0

    @property
    def end_of_document(self):
        """
        Return the offset right before \end{document}

        used by LaTeXEditor.insert_at_position
        """
        if self._end_of_document is None:
            self._end_of_document = self._find_end_of_document()
        return self._end_of_document

    def _find_end_of_packages(self):
        """
        @raise IndexError: if not found
        """
        # TODO: this should be recursive

        offset = None
        for node in self:
            if node.type == Node.COMMAND and node.value == "usepackage":
                offset = node.lastEnd
        if offset is None:
            raise IndexError
        return offset

    def _find_begin_of_preamble(self):
        # TODO: this should be recursive

        offset = 0
        for node in self:
            if node.type == Node.COMMAND and node.value == "documentclass":
                offset = node.lastEnd
        return offset

    @property
    def end_of_packages(self):
        """
        Return the offset right after the last usepackage

        used by LaTeXEditor.insert_at_position
        """
        if self._end_of_packages is None:
            try:
                self._end_of_packages = self._find_end_of_packages()
            except IndexError:
                self._end_of_packages = self._find_begin_of_preamble()

        return self._end_of_packages


class LocalizedNode(Node):
    """
    This Node type holds the start and end offsets of the substring it belongs to
    in the source
    """
    def __init__(self, type, start, end, value=None, file=None):
        Node.__init__(self, type, value)
        self.start = start
        self.end = end
        self.file = file

    @property
    def lastEnd(self):
        """
        Return the end of the last child node or of this node if
        it doesn't have children.
        """
        try:
            return self[-1].end
        except IndexError:
            return self.end
        except AttributeError:
            # AttributeError: 'Document' object has no attribute 'end'
            # FIXME: why can this happen?
            return self.end


class FatalParseException(Exception):
    """
    This raised of the Parser faces a fatal error and cannot continue
    """


from .lexer import Lexer, Token


class LaTeXParser(object):
    """
    A tree parser building an object model of nodes
    """

    # TODO: remove second parse method

    @verbose
    def _parse(self, string, documentNode, file, issue_handler):
        """
        @deprecated: use parse_string() and issues() instead
        """

        self._stack = [documentNode]

        callables = {
                Token.COMMAND : self.command,
                Token.TEXT : self.text,
                Token.BEGIN_CURLY : self.beginCurly,
                Token.END_CURLY : self.endCurly,
                Token.BEGIN_SQUARE : self.beginSquare,
                Token.END_SQUARE : self.endSquare,
                Token.COMMENT : self.comment,
                Token.VERBATIM : self.verbatim }

        try:
            for token in Lexer(string):
                callables[token.type].__call__(token.value, token.offset, file, issue_handler)
        except FatalParseException:
            return

        # check stack remainder
        for node in self._stack:
            if node.type == Node.MANDATORY_ARGUMENT or node.type == Node.EMBRACED:
                issue_handler.issue(Issue("Unclosed {", node.start, node.start + 1, file, Issue.SEVERITY_ERROR))
            elif node.type == Node.OPTIONAL_ARGUMENT:
                issue_handler.issue(Issue("Unclosed [", node.start, node.start + 1, file, Issue.SEVERITY_ERROR))

    def parse(self, string, file, issue_handler):
        """
        @param string: LaTeX source
        @param from_filename: filename from where the source is read (this is used to tag
                parts of the model)

        @rtype: Document
        """
        document_node = Document(file)
        self._parse(string, document_node, file, issue_handler)

        return document_node

    # TODO: rename methods from "command()" to "_on_command()"

    def command(self, value, offset, file, issue_handler):
        top = self._stack[-1]

        if top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = LocalizedNode(Node.COMMAND, offset, offset + len(value) + 1, value, file)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.COMMAND \
                or top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.command(value, offset, file, issue_handler)
            except IndexError:
                issue_handler.issue(Issue("Undefined Parse Error", offset, offset + 1, file, Issue.SEVERITY_ERROR))

    def text(self, value, offset, file, issue_handler):
        top = self._stack[-1]

        if top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = LocalizedNode(Node.TEXT, offset, offset + len(value), value, file)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.COMMAND:
            self._stack.pop()
            self.text(value, offset, file, issue_handler)

        elif top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.text(value, offset, file, issue_handler)
            except IndexError:
                issue_handler.issue(Issue("Undefined Parse Error", offset, offset + 1, file, Issue.SEVERITY_ERROR))
        else:
            # TODO: possible?
            issue_handler.issue(Issue("Unexpected TEXT token with %s on stack" % top.type, offset, offset + 1, file, Issue.SEVERITY_ERROR))

    def beginCurly(self, value, offset, file, issue_handler):
        top = self._stack[-1]

        if top.type == Node.COMMAND:
            node = LocalizedNode(Node.MANDATORY_ARGUMENT, offset, offset + 1, file=file)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = LocalizedNode(Node.EMBRACED, offset, offset + 1, file=file)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.beginCurly(value, offset, file, issue_handler)
            except IndexError:
                issue_handler.issue(Issue("Undefined Parse Error", offset, offset + 1, file, Issue.SEVERITY_ERROR))
        else:
            # TODO: possible?
            issue_handler.issue(Issue("Unexpected BEGIN_CURLY token with %s on stack" % top.type, offset, offset + 1, file, Issue.SEVERITY_ERROR))

    def endCurly(self, value, offset, file, issue_handler):
        try:
            # pop from stack until MANDATORY_ARGUMENT or EMBRACED
            while True:
                top = self._stack[-1]
                if top.type == Node.MANDATORY_ARGUMENT or top.type == Node.EMBRACED:
                    node = self._stack.pop()
                    break
                self._stack.pop()

            # set end offset of MANDATORY_ARGUMENT or EMBRACED
            node.end = offset + 1
        except IndexError:
            issue_handler.issue(Issue("Encountered <b>}</b> without <b>{</b>", offset, offset + 1, file, Issue.SEVERITY_ERROR))
            # we cannot continue after that
            raise FatalParseException

    def beginSquare(self, value, offset, file, issue_handler):
        top = self._stack[-1]
        if top.type == Node.COMMAND:
            node = LocalizedNode(Node.OPTIONAL_ARGUMENT, offset, offset + 1, file=file)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.TEXT:
            top.value += "["

        elif top.type == Node.MANDATORY_ARGUMENT:
            node = LocalizedNode(Node.TEXT, offset, offset + 1, "[", file)
            top.append(node)
            self._stack.append(node)

        else:
            issue_handler.issue(Issue("Unexpected BEGIN_SQUARE token with %s on stack" % top.type, offset, offset + 1, file, Issue.SEVERITY_ERROR))

    def endSquare(self, value, offset, file, issue_handler):
        try:
            node = [node for node in self._stack if node.type == Node.OPTIONAL_ARGUMENT][-1]

            # open optional argument found
            # this square closes it, so pop stack until there

            while self._stack[-1].type != Node.OPTIONAL_ARGUMENT:
                self._stack.pop()
            node = self._stack.pop()
            node.end = offset + 1

        except IndexError:
            # no open optional argument, so this "]" is TEXT

            top = self._stack[-1]
            if top.type == Node.TEXT:
                top.value += "]"

            elif top.type == Node.COMMAND:
                try:
                    self._stack.pop()
                    self.endSquare(value, offset, file, issue_handler)
                except IndexError:
                    issue_handler.issue(Issue("Undefined Parse Error", offset, offset + 1, file, Issue.SEVERITY_ERROR))

            elif top.type == Node.MANDATORY_ARGUMENT or top.type == Node.DOCUMENT or top.type == Node.OPTIONAL_ARGUMENT:
                node = LocalizedNode(Node.TEXT, offset, offset + 1, "]", file)
                top.append(node)
                self._stack.append(node)

            else:
                issue_handler.issue(Issue("Unexpected END_SQUARE token with %s on stack and no optional argument" % top.type, offset, offset + 1, file, Issue.SEVERITY_ERROR))

    _PATTERN = compile("(TODO|FIXME)\w?\:?(?P<text>.*)")

    def comment(self, value, offset, file, issue_handler):
        """
        Extract TODOs and FIXMEs
        """
        match = self._PATTERN.search(value)
        if match:
            text = match.group("text").strip()
            # TODO: escape
            issue_handler.issue(Issue(text, offset + match.start(), offset + match.end() + 1, file, Issue.SEVERITY_TASK))

    def verbatim(self, value, offset, file, issue_handler):
        pass

    #~ def __del__(self):
        #~ print "Properly destroyed %s" % self


class PrefixParser(object):
    """
    A light-weight LaTeX parser used for parsing just a prefix in
    the code completion.

    The differences between the full parser and this one include:
     * we don't collect issues (we just raise an exception)
     * we don't store node offsets
     * we indicate whether the last argument is closed or not
    """

    # TODO: use another Lexer here that doesn't count offsets (faster)

    def parse(self, string, documentNode):

        # TODO: change semantic

        self._stack = [documentNode]

        callables = {Token.COMMAND : self.command,
                     Token.TEXT : self.text,
                     Token.BEGIN_CURLY : self.beginCurly,
                     Token.END_CURLY : self.endCurly,
                     Token.BEGIN_SQUARE : self.beginSquare,
                     Token.END_SQUARE : self.endSquare,
                     Token.COMMENT : self.comment,
                     Token.VERBATIM : self.verbatim}

        try:
            for token in Lexer(string, skipWs=False, skipComment=False):
                callables[token.type].__call__(token.value)
        except FatalParseException:
            return

    def command(self, value):
        top = self._stack[-1]

        if top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = Node(Node.COMMAND, value)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.COMMAND \
                or top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.command(value)
            except IndexError:
                raise FatalParseException

    def text(self, value):
        top = self._stack[-1]

        if top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = Node(Node.TEXT, value)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.COMMAND:
            self._stack.pop()
            self.text(value)

        elif top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.text(value)
            except IndexError:
                raise FatalParseException
        else:
            # TODO: possible?
            raise FatalParseException

    def beginCurly(self, value):
        top = self._stack[-1]

        if top.type == Node.COMMAND:
            node = Node(Node.MANDATORY_ARGUMENT)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.DOCUMENT \
                or top.type == Node.MANDATORY_ARGUMENT \
                or top.type == Node.OPTIONAL_ARGUMENT \
                or top.type == Node.EMBRACED:
            node = Node(Node.EMBRACED)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.TEXT:
            try:
                self._stack.pop()
                self.beginCurly(value)
            except IndexError:
                raise FatalParseException
        else:
            # TODO: possible?
            raise FatalParseException

    def endCurly(self, value):
        try:
            # pop from stack until MANDATORY_ARGUMENT or EMBRACED
            while True:
                top = self._stack[-1]
                if top.type == Node.MANDATORY_ARGUMENT or top.type == Node.EMBRACED:
                    node = self._stack.pop()
                    break
                self._stack.pop()

            node.closed = True
        except IndexError:
            raise FatalParseException

    def beginSquare(self, value):
        top = self._stack[-1]
        if top.type == Node.COMMAND:
            node = Node(Node.OPTIONAL_ARGUMENT)
            top.append(node)
            self._stack.append(node)

        elif top.type == Node.TEXT:
            top.value += "["

        elif top.type == Node.MANDATORY_ARGUMENT:
            node = Node(Node.TEXT, "[")
            top.append(node)
            self._stack.append(node)

        else:
            raise FatalParseException

    def endSquare(self, value):
        try:
            # check whether an optional argument is open at all
            #
            # for this we address the top OPTIONAL_ARGUMENT node on the stack, if it doesn't
            # exist an IndexError is thrown
            node = [node for node in self._stack if node.type == Node.OPTIONAL_ARGUMENT][-1]

            # open optional argument found
            # this square closes it, so pop stack until there

            while self._stack[-1].type != Node.OPTIONAL_ARGUMENT:
                self._stack.pop()
            node = self._stack.pop()
            node.closed = True

        except IndexError:
            # no open optional argument, so this "]" is TEXT

            top = self._stack[-1]
            if top.type == Node.TEXT:
                top.value += "]"

            elif top.type == Node.COMMAND:
                try:
                    self._stack.pop()
                    self.endSquare(value)
                except IndexError:
                    raise FatalParseException

            elif top.type == Node.MANDATORY_ARGUMENT or top.type == Node.DOCUMENT or top.type == Node.OPTIONAL_ARGUMENT:
                node = Node(Node.TEXT, "]")
                top.append(node)
                self._stack.append(node)

            else:
                raise FatalParseException

    def comment(self, value):
        pass

    def verbatim(self, value):
        pass


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
tools

A tool is what we called a 'build profile' before. 'Tool' is more generic.
It can be used for cleaning up, converting files, for building PDFs etc.
"""

import logging

from gi.repository import Gtk, Gedit

from ..resources import Resources
from ..action import Action

LOG = logging.getLogger(__name__)

class Tool(object):
    """
    The model of a tool. This is to be stored in preferences.
    """
    def __init__(self, label, jobs, description, accelerator, extensions=[]):
        """
        @param label: a label used when displaying the Tool in the UI
        @param jobs: a list of Job objects
        @param description: a descriptive string used as tooltip
        @param accelerator: a key combination for activating this tool
        @param extensions: a list of file extensions for which this Tool can be used
        """
        self.label = label
        self.jobs = jobs
        self.description = description
        self.extensions = extensions
        self.accelerator = accelerator

    def __str__(self):
        return "Tool{%s}" % self.label


class Job(object):
    """
    A Job is one command to be executed in a Tool

    Command templates may contain the following placeholders:

     * $filename : the full filename of the processed file
     * $directory : the parent directory of the processed file
     * $shortname : the filename of the processed file without extension ('/dir/doc' for '/dir/doc.tex')
    """
    def __init__(self, command_template, must_succeed, post_processor):
        """
        Construct a Job

        @param command_template: a template string for the command to be executed
        @param must_succeed: if True this Job may cause the whole Tool to fail
        @param post_processor: a class implementing IPostProcessor
        """
        self._command_template = command_template
        self._must_succeed = must_succeed
        self._post_processor = post_processor

    @property
    def command_template(self):
        return self._command_template

    @property
    def must_succeed(self):
        return self._must_succeed

    @property
    def post_processor(self):
        return self._post_processor


class ToolAction(Action):
    """
    This hooks Tools in the UI. A ToolAction is instantiated for each registered Tool.
    """

    def __init__(self, tool):
        self._tool = tool
        self._runner = ToolRunner()

    @property
    def label(self):
        return self._tool.label

    @property
    def stock_id(self):
        return Gtk.STOCK_EXECUTE

    @property
    def accelerator(self):
        return None

    @property
    def tooltip(self):
        return self._tool.description

    def activate(self, context):
        LOG.debug("tool activate: %s" % self._tool)

        if context.active_editor:
            decorator = context._window_decorator
            doc = decorator.window.get_active_document()
            self.saving_id = doc.connect("saved",self.run_tool,context,doc)
            decorator.save_file()
        else:
            LOG.error("tool activate: no active editor")
            
    def run_tool(self, document, context, doc):
        tool_view = context.find_view(context.active_editor, "ToolView")

        self._runner.run(context.active_editor.file, self._tool, tool_view)
        LOG.debug("run tool on: %s" % context.active_editor.file)

        # destroy the save listener, or else we get a compile on any save
        doc.disconnect(self.saving_id)



from os import chdir
from .util import Process
from string import Template


class ToolRunner(Process):
    """
    This runs a Tool in a subprocess
    """

    def run(self, file, tool, issue_handler):
        """
        @param file: a File object
        @param tool: a Tool object
        @param issue_handler: an object implementing IStructuredIssueHandler
        """
        self._file = file
        self._stdout_text = ""
        self._stderr_text = ""
        self._job_iter = iter(tool.jobs)

        # add alert to the statusbar
        self._statusbar = Gedit.App.get_default().get_active_window().get_statusbar()
        self._msg_id = self._statusbar.push(1,'Compiling document ...')

        # init the IStructuredIssueHandler
        self._issue_handler = issue_handler
        self._issue_handler.clear()
        self._root_issue_partition = self._issue_handler.add_partition("<b>%s</b>" % tool.label, "running", None)
        self._issue_partitions = {}
        for job in tool.jobs:
            self._issue_partitions[job] = self._issue_handler.add_partition(job.command_template, "running", self._root_issue_partition)

        # change working directory to prevent issues with relative paths
        chdir(file.dirname)

        # enable abort
        self._issue_handler.set_abort_enabled(True, self.abort)

        # run
        self.__proceed()

    def __proceed(self):
        try:
            self._job = next(self._job_iter)

            command_template = Template(self._job.command_template)
            command = command_template.safe_substitute({"filename" : self._file.path,
                                                        "shortname" : self._file.shortname,
                                                        "directory" : self._file.dirname,
                                                        "plugin_path" : Resources().get_system_dir()})

            self._issue_handler.set_partition_state(self._issue_partitions[self._job], "running")

            self.execute(command)
        except StopIteration:
            # Tool finished successfully
            self._issue_handler.set_partition_state(self._root_issue_partition, "succeeded")
            # disable abort

            # FIXME: CRASHES
            # self._issue_handler.set_abort_enabled(False, None)

            self._on_tool_succeeded()

    def _on_stdout(self, text):
        """
        """
        LOG.debug("tool stdout: " + text)
        self._stdout_text += text

    def _on_stderr(self, text):
        """
        """
        LOG.debug("tool stderr: " + text)
        self._stderr_text += text

    def _on_abort(self):
        """
        """
        # disable abort
        self._issue_handler.set_abort_enabled(False, None)
        # mark Tool and all Jobs as aborted
        self._issue_handler.set_partition_state(self._root_issue_partition, "aborted")
        self._issue_handler.set_partition_state(self._issue_partitions[self._job], "aborted")
        for job in self._job_iter:
            self._issue_handler.set_partition_state(self._issue_partitions[job], "aborted")

    def _on_exit(self, condition):
        """
        """
        LOG.debug("tool exit")

        assert self._job

        # create post-processor instance
        post_processor = self._job.post_processor()

        # run post-processor
        post_processor.process(self._file, self._stdout_text, self._stderr_text, condition)

        # show issues
        self._issue_handler.append_issues(self._issue_partitions[self._job], post_processor.issues)

        # remove alert
        self._statusbar.remove(1,self._msg_id)

        if post_processor.successful:
            self._issue_handler.set_partition_state(self._issue_partitions[self._job], "succeeded")
            self.__proceed()
        else:
            self._issue_handler.set_partition_state(self._issue_partitions[self._job], "failed")
            if self._job.must_succeed:
                # whole Tool failed
                self._issue_handler.set_partition_state(self._root_issue_partition, "failed")
                # disable abort
                self._issue_handler.set_abort_enabled(False, None)

                self._on_tool_failed()
            else:
                self.__proceed()

    def _on_tool_succeeded(self):
        """
        The Tool has finished successfully

        To be overridden by subclass
        """

    def _on_tool_failed(self):
        """
        The Tool has failed

        To be overridden by subclass
        """

# ex:ts=4:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2011 - Ignacio Casal Quinteiro
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <http://www.gnu.org/licenses/>.

from .singleton import Singleton
import logging

LOG = logging.getLogger(__name__)


class SnippetManager(Singleton):
    def __init_once__(self):
        pass

    def insert(self, editor, iter, text):
        view = editor.tab_decorator.tab.get_view()
        window = view.get_toplevel()
        bus = window.get_message_bus()

        if bus.is_registered('/plugins/snippets', 'parse-and-activate'):
            bus.send('/plugins/snippets', 'parse-and-activate',
                     snippet=text, iter=iter, view=view)
            LOG.info("Inserted using snippets plugin")
        else:
            buf = view.get_buffer()

            buf.begin_user_action()
            buf.insert(iter, text)
            buf.end_user_action()
            LOG.info("Inserted without snippets plugin")

    def insert_at_cursor(self, editor, text):
        buf = editor.tab_decorator.tab.get_document()
        insert = buf.get_insert()
        iter = buf.get_iter_at_mark(insert)
        self.insert(editor, iter, text)

# vi:ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.model

The model of BibTeX read from an XML file.
"""
import xml.etree.ElementTree as ElementTree

from ..resources import Resources


class Type(object):
    def __init__(self, name):
        self.name = name
        self.required_fields = []
        self.optional_fields = []

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)


class Field(object):
    def __init__(self, name, label):
        self.name = name
        self.label = label


class BibTeXModel(object):
    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):
            # init object
            self._fields = {}
            self._types = {}

            # parse bibtex.xml
            self._bibtex = ElementTree.parse(Resources().get_data_file("bibtex.xml")).getroot()

            for field_e in self._bibtex.findall("fields/field"):
                id = field_e.get("id")
                self._fields[id] = Field(id, field_e.get("_label"))

            for type_e in self._bibtex.findall("types/type"):
                id = type_e.get("id")
                type = Type(id)

                for required_field_e in type_e.findall("required/field"):
                    field_id = required_field_e.get("id")
                    type.required_fields.append(self._fields[field_id])

                for optional_field_e in type_e.findall("optional/field"):
                    field_id = optional_field_e.get("id")
                    type.optional_fields.append(self._fields[field_id])

                self._types[id.lower()] = type

            self._types_list = None
            self._ready = True

    def find_type(self, name):
        """
        Find an entry type by its name
        """
        return self._types[name.lower()]

    @property
    def types(self):
        """
        List all entry types in sorted order
        """
        if self._types_list is None:
            self._types_list = list(self._types.values())
            self._types_list.sort()
        return self._types_list


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.preview
"""

from ..file import File
from ..tools import Tool, Job, ToolRunner
from ..tools.postprocess import RubberPostProcessor, GenericPostProcessor
from ..issues import MockStructuredIssueHandler
from .environment import Environment

from gi.repository import GdkPixbuf

class ImageToolGenerator(object):
    """
    This generates Tools for rendering images from LaTeX source
    """

    FORMAT_PNG, FORMAT_JPEG, FORMAT_GIF = 1, 2, 3

    PNG_MODE_MONOCHROME, PNG_MODE_GRAYSCALE, PNG_MODE_RGB, PNG_MODE_RGBA = 1, 2, 3, 4

    def __init__(self):
        self._names = {self.FORMAT_PNG : "PNG Image", self.FORMAT_JPEG : "JPEG Image", self.FORMAT_GIF : "GIF Image"}
        self._png_modes = {self.PNG_MODE_MONOCHROME : "mono", self.PNG_MODE_GRAYSCALE : "gray", self.PNG_MODE_RGB : "16m",
                        self.PNG_MODE_RGBA : "alpha"}

        # default settings
        self.format = self.FORMAT_PNG
        self.png_mode = self.PNG_MODE_RGBA
        self.render_box = True
        self.resolution = int(round(Environment().screen_dpi))
        self.antialias_factor = 4
        self.open = False

    def generate(self):
        """
        @return: a Tool object
        """
        tool = Tool(label=self._names[self.format], jobs=[], description="", accelerator="", extensions=[])

        # use rubber to render a DVI
        tool.jobs.append(Job("rubber --force --short --inplace \"$filename\"", True, RubberPostProcessor))

        if self.render_box:
            # DVI -> PS

            # -D num    resolution in DPI
            # -q        quiet mode
            # -E        generate an EPSF file with a tight bounding box
            tool.jobs.append(Job("dvips -D %s -q -E -o \"$shortname.eps\" \"$shortname.dvi\"" % self.resolution, True, GenericPostProcessor))

            # EPS -> PNG|JPG|GIF
            if self.format == self.FORMAT_PNG:
                command = "$plugin_path/util/eps2png.pl -png%s -resolution=%s -antialias=%s \"$shortname.eps\"" % (self._png_modes[self.png_mode],
                                                                                self.resolution, self.antialias_factor)
            elif self.format == self.FORMAT_JPEG:
                command = "$plugin_path/util/eps2png.pl -jpeg -resolution=%s -antialias=%s \"$shortname.eps\"" % (self.resolution, self.antialias_factor)
            elif self.format == self.FORMAT_GIF:
                command = "$plugin_path/util/eps2png.pl -gif -resolution=%s -antialias=%s \"$shortname.eps\"" % (self.resolution, self.antialias_factor)

            tool.jobs.append(Job(command, True, GenericPostProcessor))
        else:
            # dvips
            tool.jobs.append(Job("dvips -D %s -q -o \"$shortname.ps\" \"$shortname.dvi\"" % self.resolution, True, GenericPostProcessor))

            if self.format == self.FORMAT_PNG:
                tool.jobs.append(Job("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=png%s -sOutputFile=$shortname.png $shortname.ps quit.ps"
                                    % (self.resolution, self.antialias_factor, self.antialias_factor, self._png_modes[self.png_mode]), True, GenericPostProcessor))
            elif self.format == self.FORMAT_JPEG:
                tool.jobs.append(Job("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=jpeg -sOutputFile=$shortname.jpg $shortname.ps quit.ps"
                                    % (self.resolution, self.antialias_factor, self.antialias_factor), True, GenericPostProcessor))
            elif self.format == self.FORMAT_GIF:
                tool.jobs.append(Job("gs -q -dNOPAUSE -r%s -dTextAlphaBits=%s -dGraphicsAlphaBits=%s -sDEVICE=ppm -sOutputFile=$shortname.ppm $shortname.ps quit.ps"
                                    % (self.resolution, self.antialias_factor, self.antialias_factor), True, GenericPostProcessor))
                # ppmtogif
                tool.jobs.append(Job("ppmtogif $shortname.ppm > $shortname.gif", True, GenericPostProcessor))

        if self.open:
            extension = {self.FORMAT_PNG : "png", self.FORMAT_JPEG: "jpg", self.FORMAT_GIF : "gif"}[self.format]
            tool.jobs.append(Job("gvfs-open \"$shortname.%s\"" % extension, True, GenericPostProcessor))

        return tool


from tempfile import NamedTemporaryFile

class PreviewRenderer(ToolRunner):
    def render(self, source):
        """
        Render a preview image from LaTeX source

        @param source: some LaTeX source without \begin{document}
        """
        # create temp file with source
        self._temp_file = NamedTemporaryFile(mode="w", suffix=".tex")
        self._temp_file.write("\\documentclass{article}\\pagestyle{empty}\\begin{document}%s\\end{document}" % source)
        self._temp_file.flush()

        # generate Tool
        tool = ImageToolGenerator().generate()
        self._file = File(self._temp_file.name)
        issue_handler = MockStructuredIssueHandler()

        # run the Tool
        self.run(self._file, tool, issue_handler)

    def _on_tool_succeeded(self):
        # see ToolRunner._on_tool_succeeded
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(self._file.shortname + ".png")
        self.__cleanup()
        self._on_render_succeeded(pixbuf)

    def _on_tool_failed(self):
        # see ToolRunner._on_tool_failed
        self.__cleanup()
        self._on_render_failed()

    def __cleanup(self):
        """
        Remove the files created during the render process
        """
        # delete the temp file
        self._temp_file.close()

        # delete all files created during the build process
        for file in self._file.siblings:
            try:
                file.delete()
                self._log.debug("Removed %s" % file)
            except OSError:
                self._log.error("Failed to remove '%s'" % file)

    def _on_render_succeeded(self, pixbuf):
        """
        The rendering process has finished successfully

        @param pixbuf: a GdkPixbuf.Pixbuf containing the result image
        """

    def _on_render_failed(self):
        """
        The rendering process has failed
        """



# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import re
import time
import logging
import uuid

from gi.repository import Gtk, Gdk

from .completion import CompletionDistributor
from .snippetmanager import SnippetManager
from .file import File

LOG = logging.getLogger(__name__)

class Editor(object):
    """
    The base class for editors. This manages
     - the subclass life-cycle
     - the marker framework
     - change monitoring
     - drag'n'drop support
    """

    class Marker(object):
        """
        Markers refer to and highlight a range of text in the TextBuffer decorated by
        an Editor. They are used for highlighting issues.

        Each Marker instance stores two Gtk.TextMark objects refering to the start and
        end of the text range.
        """
        def __init__(self, left_mark, right_mark, id, type):
            """
            @param left_mark: a Gtk.TextMark
            @param right_mark: a Gtk.TextMark
            @param id: a unique string
            @param type: a marker type string
            """
            self.left_mark = left_mark
            self.right_mark = right_mark
            self.type = type
            self.id = id

    class MarkerTypeRecord(object):
        """
        This used for managing Marker types
        """
        def __init__(self, tag, anonymous):
            """
            @param tag: a Gtk.TextTag
            """
            self.tag = tag
            self.anonymous = anonymous
            self.markers = []

    __PATTERN_INDENT = re.compile("[ \t]+")

    # A list of file extensions
    #
    # If one or more files with one of these extensions is dragged and dropped on the editor,
    # the Editor.drag_drop_received method is called. An empty list disables the dnd support.
    dnd_extensions = []

    def __init__(self, tab_decorator, file):
        self._tab_decorator = tab_decorator
        self._file = file
        self._text_buffer = tab_decorator.tab.get_document()
        self._text_view = tab_decorator.tab.get_view()

        # hook completion handlers of the subclassing Editor
        completion_handlers = self.completion_handlers
        if len(completion_handlers):
            self._completion_distributor = CompletionDistributor(self, completion_handlers)
        else:
            self._completion_distributor = None

        #
        # init marker framework
        #

        # needed for cleanup
        self._tags = []
        self._marker_types = {}    # {marker type -> MarkerTypeRecord object}
        self._markers = {}        # { marker id -> marker object }

        self._window_context = self._tab_decorator._window_decorator._window_context
        self._window_context.create_editor_views(self, file)

        self._offset = None        # used by move_cursor

        self.__view_signal_handlers = [
                self._text_view.connect("button-press-event", self.__on_button_pressed),
                self._text_view.connect("key-release-event", self.__on_key_released),
                self._text_view.connect("button-release-event", self.__on_button_released)]

        self.__buffer_change_timestamp = time.time()
        self.__buffer_signal_handlers = [
                self._text_buffer.connect("changed", self.__on_buffer_changed)]

        # dnd support
        if len(self.dnd_extensions) > 0:
            self.__view_signal_handlers.append(
                    self._text_view.connect("drag-data-received", self.__on_drag_data_received))

        # start life-cycle for subclass
        self.init(file, self._window_context)

    def __on_buffer_changed(self, text_buffer):
        """
        Store the timestamp of the last buffer change
        """
        self.__buffer_change_timestamp = time.time()

    def __on_drag_data_received(self, widget, context, x, y, data, info, timestamp):
        """
        The drag destination received the data from the drag operation

        @param widget: the widget that received the signal
        @param context: the Gdk.DragContext
        @param x: the X position of the drop
        @param y: the Y position of the drop
        @param data: a Gtk.SelectionData object
        @param info: an integer ID for the drag
        @param timestamp: the time of the drag event
        """
        LOG.debug("drag-data-received")

        files = []
        match = False

        for uri in data.get_uris():
            file = File(uri)
            files.append(file)
            if file.extension.lower() in self.dnd_extensions:
                match = True

        if match:
            self.drag_drop_received(files)

    def __on_key_released(self, *args):
        """
        This helps to call 'move_cursor'
        """
        offset = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert()).get_offset()
        if offset != self._offset:
            self._offset = offset
            self.on_cursor_moved(offset)

    def __on_button_released(self, *args):
        """
        This helps to call 'move_cursor'
        """
        offset = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert()).get_offset()
        if offset != self._offset:
            self._offset = offset
            self.on_cursor_moved(offset)

    def __on_button_pressed(self, text_view, event):
        """
        Mouse button has been pressed on the TextView
        """
        if event.button == 3:    # right button
            x, y = text_view.get_pointer()
            x, y = text_view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET, x, y)
            it = text_view.get_iter_at_location(x, y)

            LOG.debug("Right button pressed at offset %s" % it.get_offset())

            #
            # find Marker at this position
            #
            while True:
                for mark in it.get_marks():
                    name = mark.get_name()

                    LOG.debug("Found TextMark '%s' at offset %s" % (name, it.get_offset()))

                    if name:
                        if name in list(self._markers.keys()):
                            marker = self._markers[name]
                            return self.on_marker_activated(marker, event)
                        else:
                            LOG.warning("No marker found for TextMark '%s'" % name)
                    else:
                        # FIXME: this is not safe - use another symbol for right boundaries!
                        LOG.debug("Unnamed TextMark found, outside of any Markers")
                        return

                # move left by one char and continue
                if not it.backward_char():
                    # start of buffer reached
                    return

        elif event.button == 1 and event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            x, y = text_view.get_pointer()
            x, y = text_view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET, x, y)
            it = text_view.get_iter_at_location(x, y)
            # notify subclass
            self._ctrl_left_clicked(it)

    def _ctrl_left_clicked(self, it):
        """
        Left-clicked on the editor with Ctrl modifier key pressed
        @param it: the Gtk.TextIter at the clicked position
        """

    @property
    def file(self):
        return self._file

    @property
    def tab_decorator(self):
        return self._tab_decorator

    def delete_at_cursor(self, offset):
        """
        Delete characters relative to the cursor position:

        offset < 0        delete characters from offset to cursor
        offset > 0        delete characters from cursor to offset
        """
        start = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert())
        end = self._text_buffer.get_iter_at_offset(start.get_offset() + offset)
        self._text_buffer.delete(start, end)

    # methods/properties to be used/overridden by the subclass

    @property
    def extensions(self):
        """
        Return a list of extensions for which this Editor is to be activated
        """
        raise NotImplementedError

    def drag_drop_received(self, files):
        """
        To be overridden

        @param files: a list of File objects dropped on the Editor
        """
        pass

    @property
    def initial_timestamp(self):
        """
        Return an initial reference timestamp (this just has to be smaller than
        every value returned by current_timestamp)
        """
        return 0

    @property
    def current_timestamp(self):
        """
        Return the current timestamp for buffer change recognition
        """
        return time.time()

    def content_changed(self, reference_timestamp):
        """
        Return True if the content of this Editor has changed since a given
        reference timestamp (this must be a timestamp as returned by current_timestamp)
        """
        return self.__buffer_change_timestamp > reference_timestamp

    @property
    def charset(self):
        """
        Return the character set used by this Editor
        """
        return self._text_buffer.get_encoding().get_charset()

    @property
    def content(self):
        """
        Return the string contained in the TextBuffer
        """
        return self._text_buffer.get_text(self._text_buffer.get_start_iter(),
                                    self._text_buffer.get_end_iter(), False)

    @property
    def content_at_left_of_cursor(self):
        """
        Only return the content at left of the cursor
        """
        end_iter = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert())
        return self._text_buffer.get_text(self._text_buffer.get_start_iter(),
                                    end_iter, False)

    def insert(self, source):
        """
        This may be overridden to catch special types like LaTeXSource
        """
        LOG.debug("insert(%s)" % source)

        SnippetManager().insert_at_cursor(self, str(source))

        # grab the focus again (necessary e.g. after symbol insert)
        self._text_view.grab_focus()

    def insert_at_offset(self, offset, string, scroll=False):
        """
        Insert a string at a certain offset

        @param offset: a positive int
        @param string: a str
        @param scroll: if True the view is scrolled to the insert position
        """
        iter = self._text_buffer.get_iter_at_offset(offset)
        self._text_buffer.insert(iter, str(string))

        if scroll:
            self._text_view.scroll_to_iter(iter, .25, False, 0.5, 0.5)

        # grab the focus again (necessary e.g. after symbol insert)
        self._text_view.grab_focus()

    def append(self, string):
        """
        Append some source (only makes sense with simple string) and scroll to it

        @param string: a str
        """
        self._text_buffer.insert(self._text_buffer.get_end_iter(), str(string))
        self._text_view.scroll_to_iter(self._text_buffer.get_end_iter(), .25, False, 0.5, 0.5)

        # grab the focus again (necessary e.g. after symbol insert)
        self._text_view.grab_focus()

    @property
    def indentation(self):
        """
        Return the indentation string of the line at the cursor
        """
        i_start = self._text_buffer.get_iter_at_mark(self._text_buffer.get_insert())
        i_start.set_line_offset(0)

        i_end = i_start.copy()
        i_end.forward_to_line_end()
        string = self._text_buffer.get_text(i_start, i_end, False)

        match = self.__PATTERN_INDENT.match(string)
        if match:
            return match.group()
        else:
            return ""

    def select(self, start_offset, end_offset):
        """
        Select a range of text and scroll the view to the right position
        """
        # select
        it_start = self._text_buffer.get_iter_at_offset(start_offset)
        it_end = self._text_buffer.get_iter_at_offset(end_offset)
        self._text_buffer.select_range(it_start, it_end)
        # scroll
        self._text_view.scroll_to_iter(it_end, .25, False, 0.5, 0.5)

    def select_lines(self, start_line, end_line=None):
        """
        Select a range of lines in the text

        @param start_line: the first line to select (counting from 0)
        @param end_line: the last line to select (if None only the first line is selected)
        """
        it_start = self._text_buffer.get_iter_at_line(start_line)
        if end_line:
            it_end = self._text_buffer.get_iter_at_line(end_line)
        else:
            it_end = it_start.copy()
        it_end.forward_to_line_end()
        # select
        self._text_buffer.select_range(it_start, it_end)
        # scroll
        self._text_view.scroll_to_iter(it_end, .25, False, 0.5, 0.5)

    #
    # markers are used for highlighting (anonymous)
    #

    def register_marker_type(self, marker_type, background_color, anonymous=True):
        """
        @param marker_type: a string
        @param background_color: a hex color
        @param anonymous: markers of an anonymous type may not be activated and do not get a unique ID
        """
        assert not marker_type in list(self._marker_types.keys())

        # create Gtk.TextTag
        tag = self._text_buffer.create_tag(marker_type, background=background_color)

        self._tags.append(tag)

        # create a MarkerTypeRecord for this type
        self._marker_types[marker_type] = self.MarkerTypeRecord(tag, anonymous)

    def create_marker(self, marker_type, start_offset, end_offset):
        """
        Mark a section of the text

        @param marker_type: type string
        @return: a Marker object if the type is not anonymous or None otherwise
        """

        # check offsets
        if start_offset < 0:
            LOG.error("create_marker(): start offset out of range (%s < 0)" % start_offset)
            return

        buffer_end_offset = self._text_buffer.get_end_iter().get_offset()

        if end_offset > buffer_end_offset:
            LOG.error("create_marker(): end offset out of range (%s > %s)" % (end_offset, buffer_end_offset))

        type_record = self._marker_types[marker_type]

        # hightlight
        left = self._text_buffer.get_iter_at_offset(start_offset)
        right = self._text_buffer.get_iter_at_offset(end_offset)
        self._text_buffer.apply_tag_by_name(marker_type, left, right)

        if type_record.anonymous:
            # create TextMarks
            left_mark = self._text_buffer.create_mark(None, left, True)
            right_mark = self._text_buffer.create_mark(None, right, False)

            # create Marker object
            marker = self.Marker(left_mark, right_mark, None, marker_type)

            # store Marker
            type_record.markers.append(marker)

            return None
        else:
            # create unique marker id
            id = str(uuid.uuid1())

            # create Marker object and put into map
            left_mark = self._text_buffer.create_mark(id, left, True)
            right_mark = self._text_buffer.create_mark(None, right, False)
            marker = self.Marker(left_mark, right_mark, id, marker_type)

            # store Marker
            self._markers[id] = marker
            type_record.markers.append(marker)

            return marker

    def remove_marker(self, marker):
        """
        @param marker: the Marker object to remove
        """
        # create TextIters from TextMarks
        left_iter = self._text_buffer.get_iter_at_mark(marker.left_mark)
        right_iter = self._text_buffer.get_iter_at_mark(marker.right_mark)

        # remove TextTag
        type_record = self._marker_types[marker.type]
        self._text_buffer.remove_tag(type_record.tag, left_iter, right_iter)

        # remove TextMarks
        self._text_buffer.delete_mark(marker.left_mark)
        self._text_buffer.delete_mark(marker.right_mark)

        # remove Marker from MarkerTypeRecord
        i = type_record.markers.index(marker)
        del type_record.markers[i]

        # remove from id map
        del self._markers[marker.id]

    def remove_markers(self, marker_type):
        """
        Remove all markers of a certain type
        """
        type_record = self._marker_types[marker_type]

        for marker in type_record.markers:
            assert not marker.left_mark.get_deleted()
            assert not marker.right_mark.get_deleted()

            # create TextIters from TextMarks
            left_iter = self._text_buffer.get_iter_at_mark(marker.left_mark)
            right_iter = self._text_buffer.get_iter_at_mark(marker.right_mark)

            # remove TextTag
            self._text_buffer.remove_tag(type_record.tag, left_iter, right_iter)

            # remove TextMarks
            self._text_buffer.delete_mark(marker.left_mark)
            self._text_buffer.delete_mark(marker.right_mark)

            if not type_record.anonymous:
                # remove Marker from id map
                del self._markers[marker.id]

        # remove markers from MarkerTypeRecord
        type_record.markers = []

    def replace_marker_content(self, marker, content):
        # get TextIters
        left = self._text_buffer.get_iter_at_mark(marker.left_mark)
        right = self._text_buffer.get_iter_at_mark(marker.right_mark)

        # replace
        self._text_buffer.delete(left, right)
        left = self._text_buffer.get_iter_at_mark(marker.left_mark)
        self._text_buffer.insert(left, content)

        # remove Marker
        self.remove_marker(marker)

    def on_marker_activated(self, marker, event):
        """
        A marker has been activated

        To be overridden

        @param id: id of the activated marker
        @param event: the GdkEvent of the mouse click (for raising context menus)
        """

    @property
    def completion_handlers(self):
        """
        To be overridden

        @return: a list of objects implementing CompletionHandler
        """
        return []

    def init(self, file, context):
        """
        @param file: File object
        @param context: WindowContext object
        """

    def on_save(self):
        """
        The file has been saved to its original location
        """

    def on_cursor_moved(self, offset):
        """
        The cursor has moved
        """

    def destroy(self):
        """
        The edited file has been closed or saved as another file
        """

        # disconnect signal handlers
        for handler in self.__view_signal_handlers:
            self._text_view.disconnect(handler)

        for handler in self.__buffer_signal_handlers:
            self._text_buffer.disconnect(handler)

        # delete the tags that were created for markers
        table = self._text_buffer.get_tag_table()
        for tag in self._tags:
            table.remove(tag)

        # destroy the views associated to this editor
        for i in self._window_context.editor_views[self]:
            self._window_context.editor_views[self][i].destroy()
        del self._window_context.editor_views[self]

        # unreference the tab decorator
        del self._tab_decorator

        # destroy the completion distributor
        if self._completion_distributor != None:
            self._completion_distributor.destroy()
        del self._completion_distributor

        # unreference the window context
        del self._window_context

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.listing
"""

from xml.sax import ContentHandler, parse


class Language(object):
    def __init__(self, name):
        self.name = name
        self.dialects = []


class Dialect(object):
    def __init__(self, name, default):
        self.name = name
        self.default = default


class LanguagesParser(ContentHandler):
    """
    This parses the listing.xml file containing the available languages for listings
    """
    def __init__(self):
        self._languages = None
        self._language = None

    def startElement(self, name, attributes):
        if name == "language":
            l = Language(attributes["name"])
            self._languages.append(l)
            self._language = l
        elif name == "dialect":
            default = False
            try:
                if attributes["default"] == "true":
                    default = True
            except KeyError:
                pass

            self._language.dialects.append(Dialect(attributes["name"], default))
        elif name == "no-dialect":
            self._language.dialects.append(Dialect(None, True))

    def parse(self, languages, filename):
        """
        Parse XML

        "languages" must be a list
        """
        self._languages = languages
        parse(filename, self)


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.cache
"""

from logging import getLogger

from .parser import BibTeXParser
from ..issues import MockIssueHandler


class BibTeXDocumentCache(object):
    """
    This is used to cache BibTeX document models.

    This is only used in the context of code completion and validation for LaTeX documents
    but not when editing a BibTeX file.

    Of course, this must be implemented as a singleton class.
    """

    # TODO: serialize the cache on shutdown

    _log = getLogger("BibTeXDocumentCache")

    class Entry(object):
        """
        An entry in the cache
        """
        _log = getLogger("bibtex.cache.BibTeXDocumentCache.Entry")

        def __init__(self, file):
            self.__file = file
            self.__parser = BibTeXParser(quiet=True)
            self.__issue_handler = MockIssueHandler()
            self.__mtime = 0
            self.__document = None

            self.synchronize()

        @property
        def modified(self):
            return (self.__file.mtime > self.__mtime)

        @property
        def document(self):
            return self.__document

        def synchronize(self):
            """
            Synchronize document model with file contents.

            This may throw OSError
            """
            # update timestamp
            self.__mtime = self.__file.mtime

            # parse
            self.__document = self.__parser.parse(open(self.__file.path, "r").read(), self.__file, self.__issue_handler)

    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):
            self._entries = {}
            self._ready = True

    def get_document(self, file):
        """
        Return the (hopefully) cached document model for a given file

        @param file: a File object
        """
        try:
            # update entry if necessary
            entry = self._entries[file.uri]
            if entry.modified:
                self._log.debug("File '%s' modified, synchronizing..." % file)
                entry.synchronize()
        except KeyError:
            self._log.debug("Cache fault for '%s'" % file)
            # create new entry
            entry = self.Entry(file)
            self._entries[file.uri] = entry

        return entry.document



# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
latex.environment
"""

from gi.repository import Gdk
from os import popen, system
#from Gtk.gdk import screen_width, screen_height, screen_width_mm, screen_height_mm
from pwd import getpwnam
from getpass import getuser
from locale import getdefaultlocale, nl_langinfo, D_FMT
from logging import getLogger


class CnfFile(dict):
    """
    This parses a .cnf file and provides its contents as a dictionary
    """
    def __init__(self, filename):
        """
        @raise IOError: if file is not found
        """
        for line in open(filename).readlines():
            if not line.startswith("%"):
                try:
                    key, value = line.split("=")
                    self[key.strip()] = value.strip()
                except:
                    pass


_INPUT_ENCODINGS = {
    "utf8"       : "UTF-8 (Unicode)",
    "ascii"       : "US-ASCII",
    "next"     : "ASCII (NeXT)",
    "ansinew"  : "ASCII (Windows)",
    "applemac" : "ASCII (Apple)",
    "macce"    : "MacCE (Apple Central European)",
    "latin1"   : "Latin-1 (Western European)",
    "latin2"   : "Latin-2 (Danubian European)",
    "latin3"   : "Latin-3 (South European)",
    "latin4"   : "Latin-4 (North European)",
    "latin5"   : "Latin-5 (Turkish)",
    "latin6"   : "Latin-6 (Nordic)",
    "latin7"   : "Latin-7 (Baltic)",
    "latin8"   : "Latin-8 (Celtic)",
    "latin9"   : "Latin-9 (extended Latin-1)",
    "latin10"  : "Latin-10 (South-Eastern European)",
    "cp1250"   : "CP1250 (Windows Central European)",
    "cp1252"   : "CP1252 (Windows Western European)",
    "cp1257"   : "CP1257 (Windows Baltic)",
    "cp437"    : "CP437 (DOS US with β)",
    "cp437de"  : "CP437 (DOS US with ß)",
    "cp850"    : "CP850 (DOS Latin-1)",
    "cp852"    : "CP852 (DOS Central European)",
    "cp858"    : "CP858 (DOS Western European)",
    "cp865"    : "CP865 (DOS Nordic)",
    "decmulti" : "DEC Multinational Character Set"
}

_BABEL_PACKAGES = {
    "albanian"     : "Albanian",
    "afrikaans"     : "Afrikaans", #dutch dialect
    "austrian"     : "Austrian",
    "naustrian"     : "Austrian (new spelling)",
    "bahasai"       : "Bahasa Indonesia",
    "bahasam"     : "Bahasa Malaysia",
    "basque"      : "Basque",
    "breton"     : "Breton",
    "bulgarian"     : "Bulgarian",
    "catalan"     : "Catalan",
    "croatian"     : "Croatian",
    "czech"         : "Czech",
    "danish"     : "Danish",
    "dutch"         : "Dutch",
    "australian"     : "English (AU)",
    "canadian"     : "English (CA)",
    "newzealand"     : "English (NZ)",
    "UKenglish"     : "English (UK)",
    "english"     : "English (US)",
    "esperanto"     : "Esperanto",
    "estonian"     : "Estonian",
    "finnish"     : "Finnish",
    "frenchb"     : "French",
    "acadian"     : "French (Acadian)",
    "canadien"     : "French (CA)",
    "galician"     : "Galician",
    "germanb"     : "German",
    "ngermanb"     : "German (new spelling)",
    "greek"         : "Greek",
    "polutonikogreek" : "Greek (polytonic)",
#    "athnum"     : "Greek (Athens numbering)",
    "hebrew"     : "Hebrew",
    "magyar"     : "Hungarian",
    "icelandic"     : "Icelandic",
    "interlingua"     : "Interlingua",
    "irish"         : "Irish Gaelic",
    "italian"     : "Italian",
    "latin"         : "Latin",
    "lsorbian"     : "Lower Sorbian",
    "norsk"         : "Norwegian Bokmål",
    "nynorsk"     : "Norwegian Nynorsk",
    "polish"     : "Polish",
    "portuges"     : "Portuguese (PT)",
    "brazilian"     : "Portuguese (BR)",
    "romanian"     : "Romanian",
    "russianb"     : "Russian",
    "samin"         : "North Sami",
    "scottish"     : "Scottish Gaelic",
    "serbian"     : "Serbian",
    "slovak"     : "Slovak",
    "slovene"     : "Slovene",
    "spanish"     : "Spanish",
    "swedish"     : "Swedish",
    "turkish"     : "Turkish",
    "ukraineb"     : "Ukraine",
    "usorbian"     : "Upper Sorbian",
    "welsh"         : "Welsh"
}

_DOCUMENT_CLASSES = {
    "abstbook"    : _("Book of abstracts"),
    "article"     : _("Article"),
    "amsart"    : _("Article (AMS)"),
    "amsbook"    : _("Book (AMS)"),
    "amsdtx"    : _("AMS Documentation"),
    "amsproc"    : _("Proceedings (AMS)"),
    "report"     : _("Report"),
    "beamer"     : _("Beamer slides"),
    "beletter"    : _("Belgian letter"),
    "book"         : _("Book"),
    "flashcard"    : _("Flashcard"),
    "iagproc"    : _("Proceedings (IAG)"),
    "letter"     : _("Letter"),
    "ltnews"    : _("LaTeX News"),
    "ltxdoc"    : _("LaTeX Documentation"),
    "ltxguide"    : _("LaTeX Guide"),
    "proc"        : _("Proceedings"),
    "scrartcl"     : _("Article (KOMA-Script)"),
    "scrreport"    : _("Report (KOMA-Script)"),
    "scrbook"     : _("Book (KOMA-Script)"),
    "scrlettr"     : _("Letter (KOMA-Script)"),
    "scrlttr2"     : _("Letter 2 (KOMA-Script)"),
    "scrreprt"    : _("Report (KOMA-Script)"),
    "slides"    : _("Slides")
}


class TeXResource(object):
    def __init__(self, file, name, label):
        """
        @param file: a File object
        @param name: the identifier of this resource (e.g. 'ams' for 'ams.bib')
        @param label: a descriptive label
        """
        self.file = file
        self.name = name
        self.label = label


from os.path import expanduser

from ..file import File


class Environment(object):

    _CONFIG_FILENAME = "/etc/texmf/texmf.cnf"
    _DEFAULT_TEXMF_DIR = "/usr/share/texmf-texlive"
    _DEFAULT_TEXMF_DIR_HOME = "~/texmf"

    """
    This encapsulates the user's LaTeX distribution and provides methods
    for searching it
    """

    _log = getLogger("Environment")

    def __new__(cls):
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        if not '_ready' in dir(self):
            self._bibtex_styles = None
            self._classes = None
            self._language_definitions = None
            self._input_encodings = None
            self._screen_dpi = None
            self._kpsewhich_checked = False
            self._kpsewhich_installed = None
            self._file_exists_cache = {}

            self._search_paths = []
            default_search_paths = [self._DEFAULT_TEXMF_DIR, expanduser(self._DEFAULT_TEXMF_DIR_HOME)]

            try:
                cnf_file = CnfFile(self._CONFIG_FILENAME)

                path_found = False

                for key in ["TEXMFMAIN", "TEXMFDIST", "TEXMFHOME"]:
                    try:
                        self._search_paths.append(cnf_file[key])
                        path_found = True
                    except KeyError:
                        # config key not found
                        self._log.error("Key %s not found in %s" % (key, self._CONFIG_FILENAME))

                if not path_found:
                    self._log.error("No search paths found in %s, using default search paths %s" % (self._CONFIG_FILENAME, default_search_paths))
                    self._search_paths = default_search_paths

            except IOError:
                # file _CONFIG_FILENAME not found - use default path
                self._log.error("%s not found, using default search paths %s" % (self._CONFIG_FILENAME, default_search_paths))
                self._search_paths = default_search_paths

            self._ready = True

    @property
    def kpsewhich_installed(self):
        """
        Return whether kpsewhich is installed
        """
        if not self._kpsewhich_checked:
            self._kpsewhich_installed = bool(system("kpsewhich --version $2>/dev/null") == 0)
            self._kpsewhich_checked = True
        return self._kpsewhich_installed

    def file_exists(self, filename):
        """
        Uses kpsewhich to check if a TeX related file (.bst, .sty etc.) exists. The result
        is cached to minimize 'kpsewhich' calls.
        """
        if not self.kpsewhich_installed:
            return True

        try:
            return self._file_exists_cache[filename]
        except KeyError:
            found = popen("kpsewhich %s" % filename).read().splitlines()
            exists = bool(len(found))
            self._file_exists_cache[filename] = exists
            return exists

    @property
    def bibtex_styles(self):
        """
        Return the available .bst files
        """
        if not self._bibtex_styles:
            self._bibtex_styles = self._find_resources("", ".bst", {})
        return self._bibtex_styles

    @property
    def document_classes(self):
        """
        Return the available document classes
        """
        if not self._classes:
            self._classes = self._find_resources("", ".cls", _DOCUMENT_CLASSES)
        return self._classes

    @property
    def language_definitions(self):
        if not self._language_definitions:
            self._language_definitions = self._find_resources("/tex/generic/babel/", ".ldf", _BABEL_PACKAGES)
        return self._language_definitions

    @property
    def input_encodings(self):
        """
        Return a list of all available input encodings
        """
        if not self._input_encodings:
            self._input_encodings = self._find_resources("/tex/latex/base/", ".def", _INPUT_ENCODINGS)
        return self._input_encodings

    def _find_resources(self, relative, extension, labels):
        """
        Find TeX resources

        @param relative: a path relative to TEXMF... search path, e.g. '/tex/latex/base/'
        @param extension: the file extension of the resources, e.g. '.bst'
        @param labels: the dictionary to be searched for labels
        """
        resources = []
        files = []

        for search_path in self._search_paths:
            files += [File(f) for f in popen("find %s%s -name '*%s'" % (search_path, relative, extension)).readlines()]

        if len(files) > 0:
            for file in files:
                name = file.shortbasename
                try:
                    label = labels[name]
                except KeyError:
                    label = ""
                resources.append(TeXResource(file, name, label))
        else:
            # no files found
            self._log.error("No %s-files found in %s%s" % (extension, search_path, relative))

        for name, label in labels.items():
            found = False
            for resource in resources:
                if resource.name == name:
                    found = True
            if not found:
                resources.append(TeXResource(None, name, label))

        return resources

    @property
    def screen_dpi(self):
        if not self._screen_dpi:
            screen = Gdk.Screen.get_default()
            dpi_x = screen.width() / screen.width_mm() * 25.4
            dpi_y = screen.height() / screen.height_mm() * 25.4

            self._screen_dpi = (dpi_x + dpi_y) / 2.0

        return self._screen_dpi

    @property
    def username(self):
        """
        Return user name derived from pwd entry
        """
        record = getpwnam(getuser()) # get pwd entry

        self._log.debug("Found user pw entry: " + str(record))

        if len(record[4]):
            return record[4].split(",")[0]
        else:
            return record[0].title()

    @property
    def date_format(self):
        """
        Return localized date format for use in strftime()
        """
        return nl_langinfo(D_FMT)

    @property
    def language_code(self):
        """
        Return language code like 'de'
        """
        return getdefaultlocale()[0]


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
base

These classes form the interface exposed by the plugin base layer.
"""

import logging

from .file import File

LOG = logging.getLogger(__name__)


class WindowContext(object):
    """
    The WindowContext is passed to Editors and is used to
     * retrieve View instances
     * activate a specific Editor instance
     * retrieve the currently active Editor

    This also creates and destroys the View instances.
    """

    def __init__(self, window_decorator, editor_view_classes):
        """
        @param window_decorator: the GeditWindowDecorator this context corresponds to
        @param editor_view_classes: a map from extension to list of View classes
        """
        self._window_decorator = window_decorator
        self._editor_view_classes = editor_view_classes

        self.editor_views = {}    # maps Editor object to a map from ID to View object

    def create_editor_views(self, editor, file):
        """
        Create instances of the editor specific Views for a given Editor instance
        and File

        Called by Editor base class
        """
        self.editor_views[editor] = {}
        try:
            for id, clazz in self._editor_view_classes[file.extension].items():
                # create View instance and add it to the map
                self.editor_views[editor][id] = clazz(self, editor)

                LOG.debug("Created view " + id)
        except KeyError:
            LOG.debug("No views for %s" % file.extension)

    ###
    # public interface

    @property
    def active_editor(self):
        """
        Return the active Editor instance
        """
        return self._window_decorator._active_tab_decorator.editor

    def activate_editor(self, file):
        """
        Activate the Editor containing a given File or open a new tab for it

        @param file: a File object

        @raise AssertError: if the file is no File object
        """
        assert type(file) is File

        self._window_decorator.activate_tab(file)

    def find_view(self, editor, view_id):
        """
        Return a View object
        """
        try:
            return self.editor_views[editor][view_id]
        except KeyError:
            LOG.critical("Unknown view id: %s (we have: %s)" % (
                    view_id,
                    ",".join(list(self.editor_views.get(editor,{}).keys()))))

    def set_action_enabled(self, action_id, enabled):
        """
        Enable/disable an IAction object
        """
        self._window_decorator._action_group.get_action(action_id).set_sensitive(enabled)

    def destroy(self):
        # unreference the window decorator
        del self._window_decorator


# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2010 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

"""
bibtex.editor
"""

BENCHMARK = True

import logging
if BENCHMARK: import time

from ..editor import Editor
from ..preferences import Preferences
from ..issues import Issue, IIssueHandler, MockIssueHandler
from ..util import verbose

from ..job import Job, JobChangeListener

from .parser import BibTeXParser
from .completion import BibTeXCompletionHandler
from .validator import BibTeXValidator

LOG = logging.getLogger(__name__)

class ParseJob(Job):
    def _run(self, arguments):
        file = arguments[0]
        content = arguments[1]
        self._parser = BibTeXParser()
        return self._parser.parse(content, file, MockIssueHandler())


class BibTeXEditor(Editor, IIssueHandler, JobChangeListener):

    extensions = [".bib"]

    @property
    def completion_handlers(self):
        self.__bibtex_completion_handler = BibTeXCompletionHandler()
        return [self.__bibtex_completion_handler]

    def __init__(self, tab_decorator, file):
        Editor.__init__(self, tab_decorator, file)
        self._parse_job = None

    def init(self, file, context):
        LOG.debug("init(%s)" % file)

        self._preferences = Preferences()

        self._file = file
        self._context = context

        style_scheme = self._text_buffer.get_style_scheme()
        w_style = style_scheme.get_style('def:warning')
        w_color = w_style.get_properties('background')[0]
        self.register_marker_type("bibtex-warning", w_color)

        e_style = style_scheme.get_style('def:error')
        e_color = e_style.get_properties('background')[0]
        self.register_marker_type("bibtex-error", e_color)

        self._issue_view = context.find_view(self, "IssueView")
        self._parser = BibTeXParser()
        self._validator = BibTeXValidator()
        self._outline_view = context.find_view(self, "BibTeXOutlineView")

        self._parse_job = ParseJob()
        self._parse_job.set_change_listener(self)

        # initially parse
        self.__parse()

    def on_save(self):
        """
        The file has been saved

        Update models
        """
        self.__parse()

#    def _on_state_changed(self, state):
#        #
#        # job.JobChangeListener._on_state_changed
#        #
#        if (state == JobManager.STATE_COMPLETED):
#            self._log.debug("Parser finished")
#
#            self._document = self._parse_job.get_returned()
#
#            #self._log.debug(str(self._document))
#
#            for entry in self._document.entries:
#                print entry.key
#
#            # FIXME: gedit crashes here with:
#            # gedit: Fatal IO error 11 (Resource temporarily unavailable) on X server :0.0.
#
#            self._log.debug("Validating...")
#            self._validator.validate(self._document, self._file, self)
#            self._log.debug("Validating finished")
#
#            self._outline_view.set_outline(self._document)
#
#
#    def __parse_async(self):
#
#        # reset highlight
#        self.remove_markers("bibtex-error")
#        self.remove_markers("bibtex-warning")
#
#        # reset issues
#        self._issue_view.clear()
#
#        self._log.debug("Starting parser job")
#
#        self._parse_job.set_argument([self._file, self.content])
#        self._parse_job.schedule()

    @verbose
    def __parse(self):
        """
        """
        LOG.debug("__parse")

        content = self.content

        # reset highlight
        self.remove_markers("bibtex-error")
        self.remove_markers("bibtex-warning")

        # reset issues
        self._issue_view.clear()

#        self.parse(self._file)

        if BENCHMARK:
            t = time.clock()

        # parse document
        self._document = self._parser.parse(content, self._file, self)

        if BENCHMARK:
            LOG.info("BibTeXParser.parse: %f" % (time.clock() - t))

        LOG.debug("Parsed %s bytes of content" % len(content))

        # validate
        if BENCHMARK:
            t = time.clock()

        self._validator.validate(self._document, self._file, self)

        # 0.11
        if BENCHMARK:
            LOG.info("BibTeXValidator.validate: %f" % (time.clock() - t))

        self._outline_view.set_outline(self._document)

#    def _on_parser_finished(self, model):
#        """
#        """
#        self._document = model
#        self._outline_view.set_outline(self._document)

    def issue(self, issue):
        # overriding IIssueHandler.issue

        self._issue_view.append_issue(issue)

        if issue.file == self._file:
            if issue.severity == Issue.SEVERITY_ERROR:
                self.create_marker("bibtex-error", issue.start, issue.end)
            elif issue.severity == Issue.SEVERITY_WARNING:
                self.create_marker("bibtex-warning", issue.start, issue.end)

    def on_cursor_moved(self, offset):
        """
        The cursor has moved
        """
        if self._preferences.get("outline-connect-to-editor"):
            self._outline_view.select_path_by_offset(offset)

    def destroy(self):
        # unreference the window context
        del self._context

        # remove parse listener
        if self._parse_job != None:
            self._parse_job.set_change_listener(None)

        Editor.destroy(self)

# ex:ts=4:et:



# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2011 Ignacio Casal Quinteiro
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

class Singleton(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(
                     cls, *args, **kwargs)
            cls._instance.__init_once__()

        return cls._instance
